#!/usr/bin/perl
#Copyright Massachusetts Institute of technology, 2003.
#Written by Eric Prud'hommeaux for the World Wide Web Consortium
#This file is PUBLIC DOMAIN.

# AlgaeCompileTree.pm -- Object definitions for populating an algae parse tree.
# 
# $Id: AlgaeCompileTree.pm,v 1.83 2004/06/18 10:46:33 eric Exp $

use strict;
package W3C::Rdf::AlgaeCompileTree;
use W3C::Util::Exception;
require Exporter;
@W3C::Rdf::AlgaeCompileTree::ISA = qw(Exporter);
use W3C::Rdf::Algae2 qw(); # &getTripleAttribute &setTripleAttribute);
use vars qw(@EXPORT_OK
	    $DISJUNCTION $NEGATION $OUTER
	    $DefaultPrecedence $MAX_PREC $LEFT $RIGHT);
@EXPORT_OK = qw($DISJUNCTION $NEGATION $OUTER $DefaultPrecedence $MAX_PREC $LEFT $RIGHT);

($DISJUNCTION, $NEGATION, $OUTER) = qw(8 13 23);
($LEFT, $RIGHT) = (1, 2);
$MAX_PREC = 9;
my $MAX_PREC = $MAX_PREC;
$DefaultPrecedence = {
    '||' => [9, $LEFT],
    '^' => [8, $LEFT],
    '.' => [7, $LEFT],
    '&&' => [7, $LEFT],
    '<' => [5, $LEFT],
    '<=' => [5, $LEFT],
    '>' => [5, $LEFT],
    '>=' => [5, $LEFT],
    '==' => [5, $LEFT],
    '+' => [3, $RIGHT],
    '-' => [3, $RIGHT],
    '*' => [2, $LEFT],
    '/' => [2, $LEFT],
    '!' => [1, $RIGHT],
    NEG => [1, $RIGHT],
};
my $DefaultPrecedence = $DefaultPrecedence;



# Each class corresponds to a production in the AlgaeParser grammar.

#                    Action
#                    /    \
#        DelayedAction    ImmediateAction _____________________________
#           /     \        |    \     \       \    \     \             \
# TriplesAction  FwRule  Attach Clear Collect Host Flags NamespaceDecl Slurp
#   /     \
# Ask    Assert
@W3C::Rdf::AlgaeCompileTree::ImmediateAction::ISA = qw(W3C::Rdf::AlgaeCompileTree::Action);
@W3C::Rdf::AlgaeCompileTree::DelayedAction::ISA = qw(W3C::Rdf::AlgaeCompileTree::Action);
@W3C::Rdf::AlgaeCompileTree::TriplesAction::ISA = qw(W3C::Rdf::AlgaeCompileTree::DelayedAction);

@W3C::Rdf::AlgaeCompileTree::Ask::ISA = qw(W3C::Rdf::AlgaeCompileTree::TriplesAction);
@W3C::Rdf::AlgaeCompileTree::Assert::ISA = qw(W3C::Rdf::AlgaeCompileTree::TriplesAction);
@W3C::Rdf::AlgaeCompileTree::Attach::ISA = qw(W3C::Rdf::AlgaeCompileTree::ImmediateAction);
@W3C::Rdf::AlgaeCompileTree::Clear::ISA = qw(W3C::Rdf::AlgaeCompileTree::ImmediateAction);
@W3C::Rdf::AlgaeCompileTree::CollectSet::ISA = qw(W3C::Rdf::AlgaeCompileTree::ImmediateAction);
@W3C::Rdf::AlgaeCompileTree::Collect::ISA = qw(W3C::Rdf::AlgaeCompileTree::CollectSet);
@W3C::Rdf::AlgaeCompileTree::CollectStar::ISA = qw(W3C::Rdf::AlgaeCompileTree::CollectSet);
@W3C::Rdf::AlgaeCompileTree::Host::ISA = qw(W3C::Rdf::AlgaeCompileTree::ImmediateAction);
@W3C::Rdf::AlgaeCompileTree::Flags::ISA = qw(W3C::Rdf::AlgaeCompileTree::ImmediateAction);
@W3C::Rdf::AlgaeCompileTree::FwRule::ISA = qw(W3C::Rdf::AlgaeCompileTree::DelayedAction);
@W3C::Rdf::AlgaeCompileTree::NamespaceDecl::ISA = qw(W3C::Rdf::AlgaeCompileTree::ImmediateAction);
@W3C::Rdf::AlgaeCompileTree::Slurp::ISA = qw(W3C::Rdf::AlgaeCompileTree::ImmediateAction);

# Other Things
@W3C::Rdf::AlgaeCompileTree::Binding::ISA = qw(W3C::Rdf::AlgaeCompileTree::Thing);

#     ____________Thing
#    /          /       \
# Decl   UnaryDecl      BinaryDecl
#        /    \          /       \
# Negation  Option  Conjunction  Disjunction
@W3C::Rdf::AlgaeCompileTree::Decl::ISA = qw(W3C::Rdf::AlgaeCompileTree::Thing);
@W3C::Rdf::AlgaeCompileTree::UnaryDecl::ISA = qw(W3C::Rdf::AlgaeCompileTree::Thing);
@W3C::Rdf::AlgaeCompileTree::BinaryDecl::ISA = qw(W3C::Rdf::AlgaeCompileTree::Thing);
@W3C::Rdf::AlgaeCompileTree::Negation::ISA = qw(W3C::Rdf::AlgaeCompileTree::UnaryDecl);
@W3C::Rdf::AlgaeCompileTree::Option::ISA = qw(W3C::Rdf::AlgaeCompileTree::UnaryDecl);
@W3C::Rdf::AlgaeCompileTree::Conjunction::ISA = qw(W3C::Rdf::AlgaeCompileTree::BinaryDecl);
@W3C::Rdf::AlgaeCompileTree::ShortcutDisjunction::ISA = qw(W3C::Rdf::AlgaeCompileTree::BinaryDecl);
@W3C::Rdf::AlgaeCompileTree::UnionDisjunction::ISA = qw(W3C::Rdf::AlgaeCompileTree::BinaryDecl);
@W3C::Rdf::AlgaeCompileTree::UnionDisjunctionMerge::ISA = qw(W3C::Rdf::AlgaeCompileTree::BinaryDecl);
#@W3C::Rdf::AlgaeCompileTree::Disjunction::ISA = qw(W3C::Rdf::AlgaeCompileTree::ShortcutDisjunction);

#         _Thing_____
#        /           \
#     POS             StaticPOS_
#    /   \           /    \     \
#  Var  KeyName  BNode  Literal  UrlBase
#                           |    /   \
#                          NUM  URL  QName
@W3C::Rdf::AlgaeCompileTree::POS::ISA = qw(W3C::Rdf::AlgaeCompileTree::Thing);
@W3C::Rdf::AlgaeCompileTree::StaticPOS::ISA = qw(W3C::Rdf::AlgaeCompileTree::Thing);
@W3C::Rdf::AlgaeCompileTree::Var::ISA = qw(W3C::Rdf::AlgaeCompileTree::POS);
@W3C::Rdf::AlgaeCompileTree::KeyName::ISA = qw(W3C::Rdf::AlgaeCompileTree::POS);
@W3C::Rdf::AlgaeCompileTree::BNode::ISA = qw(W3C::Rdf::AlgaeCompileTree::StaticPOS);
@W3C::Rdf::AlgaeCompileTree::Literal::ISA = qw(W3C::Rdf::AlgaeCompileTree::StaticPOS);
@W3C::Rdf::AlgaeCompileTree::UrlBase::ISA = qw(W3C::Rdf::AlgaeCompileTree::StaticPOS);
@W3C::Rdf::AlgaeCompileTree::Num::ISA = qw(W3C::Rdf::AlgaeCompileTree::Literal);
@W3C::Rdf::AlgaeCompileTree::Int::ISA = qw(W3C::Rdf::AlgaeCompileTree::Num);
@W3C::Rdf::AlgaeCompileTree::Float::ISA = qw(W3C::Rdf::AlgaeCompileTree::Num);
@W3C::Rdf::AlgaeCompileTree::Url::ISA = qw(W3C::Rdf::AlgaeCompileTree::UrlBase);
@W3C::Rdf::AlgaeCompileTree::DelayedQName::ISA = qw(W3C::Rdf::AlgaeCompileTree::UrlBase);
@W3C::Rdf::AlgaeCompileTree::QName::ISA = qw(W3C::Rdf::AlgaeCompileTree::DelayedQName);

# <CONSTRAINTS>
#         Thing
#         /   \
# Constraint  Expr
#            /  \
#       Unary  Binary
#      /  \     |
#    Neg  Not  Eq,Ne,Glob,Or,Xor,And,Lt,Le,Gt,Ge,Plus,Minus,Multiply,Exp,Divide,Assign
@W3C::Rdf::AlgaeCompileTree::Constraint::ISA = qw(W3C::Rdf::AlgaeCompileTree::Thing);
@W3C::Rdf::AlgaeCompileTree::Expr::ISA = qw(W3C::Rdf::AlgaeCompileTree::Thing);

@W3C::Rdf::AlgaeCompileTree::Unary::ISA = qw(W3C::Rdf::AlgaeCompileTree::Expr);
@W3C::Rdf::AlgaeCompileTree::Neg::ISA = qw(W3C::Rdf::AlgaeCompileTree::Unary);
@W3C::Rdf::AlgaeCompileTree::Not::ISA = qw(W3C::Rdf::AlgaeCompileTree::Unary);
@W3C::Rdf::AlgaeCompileTree::BitFlip::ISA = qw(W3C::Rdf::AlgaeCompileTree::Unary);
@W3C::Rdf::AlgaeCompileTree::Lang::ISA = qw(W3C::Rdf::AlgaeCompileTree::Unary);
@W3C::Rdf::AlgaeCompileTree::Label::ISA = qw(W3C::Rdf::AlgaeCompileTree::Unary);

@W3C::Rdf::AlgaeCompileTree::Binary::ISA = qw(W3C::Rdf::AlgaeCompileTree::Expr);
@W3C::Rdf::AlgaeCompileTree::BinaryTest::ISA = qw(W3C::Rdf::AlgaeCompileTree::Binary);
@W3C::Rdf::AlgaeCompileTree::Eq::ISA = qw(W3C::Rdf::AlgaeCompileTree::BinaryTest);
@W3C::Rdf::AlgaeCompileTree::Ne::ISA = qw(W3C::Rdf::AlgaeCompileTree::BinaryTest);
@W3C::Rdf::AlgaeCompileTree::Glob::ISA = qw(W3C::Rdf::AlgaeCompileTree::Binary);
@W3C::Rdf::AlgaeCompileTree::Or::ISA = qw(W3C::Rdf::AlgaeCompileTree::Binary);
@W3C::Rdf::AlgaeCompileTree::Xor::ISA = qw(W3C::Rdf::AlgaeCompileTree::Binary);
@W3C::Rdf::AlgaeCompileTree::And::ISA = qw(W3C::Rdf::AlgaeCompileTree::Binary);
@W3C::Rdf::AlgaeCompileTree::Lt::ISA = qw(W3C::Rdf::AlgaeCompileTree::BinaryTest);
@W3C::Rdf::AlgaeCompileTree::Le::ISA = qw(W3C::Rdf::AlgaeCompileTree::BinaryTest);
@W3C::Rdf::AlgaeCompileTree::Gt::ISA = qw(W3C::Rdf::AlgaeCompileTree::BinaryTest);
@W3C::Rdf::AlgaeCompileTree::Ge::ISA = qw(W3C::Rdf::AlgaeCompileTree::BinaryTest);
@W3C::Rdf::AlgaeCompileTree::Plus::ISA = qw(W3C::Rdf::AlgaeCompileTree::Binary);
@W3C::Rdf::AlgaeCompileTree::Minus::ISA = qw(W3C::Rdf::AlgaeCompileTree::Binary);
@W3C::Rdf::AlgaeCompileTree::Multiply::ISA = qw(W3C::Rdf::AlgaeCompileTree::Binary);
@W3C::Rdf::AlgaeCompileTree::Exp::ISA = qw(W3C::Rdf::AlgaeCompileTree::Binary);
@W3C::Rdf::AlgaeCompileTree::Divide::ISA = qw(W3C::Rdf::AlgaeCompileTree::Binary);
@W3C::Rdf::AlgaeCompileTree::Modulo::ISA = qw(W3C::Rdf::AlgaeCompileTree::Binary);
@W3C::Rdf::AlgaeCompileTree::Assign::ISA = qw(W3C::Rdf::AlgaeCompileTree::Binary);
# </CONSTRAINTS>

sub format {
    my ($in, $add) = @_;
    my @lines = split("\n", $in);
    return join ("\n", $lines[0], map {"$add$_"} @lines[1..@lines-1]);
}
sub main::_defaultPrec {
    my ($ob, %flags) = @_;
    if (!exists $flags{-parentPrec}) {
	$flags{-parentPrec} = $MAX_PREC;
    }
    if (!exists $flags{-precTable}) {
	$flags{-precTable} = $DefaultPrecedence;
	$flags{-parentPrec} = $MAX_PREC+1;
    }
    return ($ob, %flags);
}

# EvalFailure Exceptions
# These exceptions simply imply that the conjunction in which they were called
# fails. Eg, adding 5 to a bNode.

@W3C::Rdf::SafeEvaluationException::ISA = qw(W3C::Util::Exception);

package W3C::Rdf::FailureOnTypeException;
@W3C::Rdf::FailureOnTypeException::ISA = qw(W3C::Rdf::SafeEvaluationException);
sub new {
    my ($proto, @parms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@parms);
    $self->missingParm('-expect') if (!exists $self->{-expect});
    $self->missingParm('-got') if (!exists $self->{-got});
    $self->fillInStackTrace;
    return $self;
}

sub getSpecificMessage {return 'Safe failure: expected: '.$_[0]->getExpect.' got: '.$_[0]->getGot();}
sub getExpect {return $_[0]->{-expect}}
sub getGot {return $_[0]->{-got}}


# The whole algae request structure.

package W3C::Rdf::AlgaeCompileTree::Action;
use W3C::Util::Exception;
sub new {
    my ($proto, @actionParms) = @_;
    my $class = ref($proto) || $proto;
    $actionParms[0] || &throw();
    my $self = {ALGAE2 => $actionParms[0]};
    bless ($self, $class);
    return $self;
}
sub toString {&throw(new W3C::Util::MethodNotImplementedException(-object => $_[0], -function => 'W3C::Rdf::AlgaeCompileTree::Action::toString'))}

package W3C::Rdf::AlgaeCompileTree::ImmediateAction;
use W3C::Util::Exception;
sub immediateEvaluate {&throw(new W3C::Util::NotImplementedException(-object => $_[0]))}
sub delayedEvaluate {}

package W3C::Rdf::AlgaeCompileTree::DelayedAction;
use W3C::Util::Exception;
sub immediateEvaluate {}
sub delayedEvaluate {&throw(new W3C::Util::NotImplementedException(-object => $_[0]))}

package W3C::Rdf::AlgaeCompileTree::Attach;
use W3C::Util::Exception;
sub new {
    my ($proto, $type, $name, $parms, @actionParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@actionParms);
    $self->{TYPE} = $type;
    $self->{PARMS} = $parms;
    $self->{NAME} = $name;
    my $typeStr = $type->getUrl;
    # some default invocation parameters:
    if ($typeStr eq 'http://www.w3.org/1999/02/26-modules/algae#dynamic') {
	my $class = delete $parms->{-class};
	if (!$class) {
	    &throw(new W3C::Util::Exception(-message => "no class name specified for dynamic database"));
	}
	$class = $class->getString;
	my $db;
	#print "require $class; \$db = new $class(\$parms);\n";
	eval "require $class; \$db = new $class(\$parms);";
	if ($@) {if (my $ex = &catch('W3C::Util::Exception')) {
	    &throw($ex);
	    #$self->{-errorHandler}->warning($ex);
	} else {
	    &throw(new W3C::Util::Exception(-message => "got exception \"$@\" while executing \"require $class; \$db = new $class(\$parms);\"."));
	    #$self->{-errorHandler}->error(new W3C::Util::PerlException);
	}}
	$self->{DB} = $db;
    } elsif ($typeStr eq 'http://www.w3.org/2003/01/21-RDF-RDB-access/ns#dbs') {
	require W3C::Rdf::SqlDB;
	foreach my $binding (@{$self->{BINDINGS}}) {
	    my ($attr, $value) = ($binding->{NAME}, $binding->{VALUE});
	    $parms->{'-'.$attr} = $value;
	}
	$self->{DB} = new W3C::Rdf::SqlDB($parms);
    } elsif ($typeStr eq 'http://www.w3.org/1999/02/26-modules/algae#sql') {
	require W3C::Rdf::SqlDB;
	foreach my $binding (@{$self->{BINDINGS}}) {
	    my ($attr, $value) = ($binding->{NAME}, $binding->{VALUE});
	    $parms->{'-'.$attr} = $value;
	}
	$self->{DB} = new W3C::Rdf::SqlDB($parms);
    } elsif ($typeStr eq 'http://www.w3.org/1999/02/26-modules/algae#ephemeral') {
	require W3C::Rdf::RdfDB;
	foreach my $binding (@{$self->{BINDINGS}}) {
	    my ($attr, $value) = ($binding->{NAME}, $binding->{VALUE});
	    $parms->{'-'.$attr} = $value;
	}
	$self->{DB} = new W3C::Rdf::RdfDB($parms);
    } else {
	&throw(new W3C::Util::Exception(-message => "unknown database \"$typeStr\""));
    }

    $self->{ALGAE2}->addSource($name, $self->{DB});
    return $self;
}
sub immediateEvaluate {
    my ($self, $resultSet) = @_;
}
sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    return "attach '(\"$self->{TYPE}\" (".join (' ', map {$_->toString} @{$self->{BINDINGS}}).'))';
}
package W3C::Rdf::AlgaeCompileTree::Binding;
sub new {
    my ($proto, $name, $value, @actionParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@actionParms);
    $self->{NAME} = $name;
    $self->{VALUE} = $value;
    return $self;
}
sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    return $self->{NAME}.'='.($self->{VALUE}); # ->toString
}
package W3C::Rdf::AlgaeCompileTree::NamespaceDecl;
sub new {
    my ($proto, $prefix, $namespace, @actionParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@actionParms);
    $self->{PREFIX} = $prefix;
    $self->{NAMESPACE} = $namespace;
    return $self;
}
sub immediateEvaluate {
    my ($self, $resultSet) = @_;
    $self->{ALGAE2}{-namespaceHandler}->addNamespace($self->{PREFIX}, $self->{NAMESPACE});
}
sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    my $prefixStr = $self->{PREFIX} ? "$self->{PREFIX}=" : '';
    return "namespace $prefixStr<$self->{NAMESPACE}>";
}
package W3C::Rdf::AlgaeCompileTree::Flags;
use W3C::Rdf::Atoms qw($ATTRIB_GroundFact);
sub new {
    my ($proto, $bindings, $setOp, @actionParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@actionParms);
    $self->{BINDINGS} = $bindings;
    $self->{SET_OP} = $setOp;
    return $self;
}
sub immediateEvaluate {
    my ($self, $resultSet) = @_;
    my $defaultParms = $self->{ALGAE2}->getDefaultParms;
    if ($self->{SET_OP} eq '') {
	foreach my $key (keys %$defaultParms) {
	    delete $defaultParms->{$key};
	}
    }
    foreach my $binding (@{$self->{BINDINGS}}) {
	my ($attr, $value) = ($binding->{NAME}, $binding->{VALUE});
	if ($attr->isa('W3C::Rdf::AlgaeCompileTree::KeyName')) {
	    my $keyName = $attr->getName;
	    if ($keyName eq 'ATTRIB') {
		my $auth = undef; # !!!
		$attr = '-sourceAttribution';
		$value = $self->{ALGAE2}{-atomDictionary}->getGroundFactAttribution(
							    $value->interned, 
							    undef, $auth, undef);
	    } else {
		&throw(new W3C::Util::Exception(-message =>"unknown key word \"$keyName}\""));
	    }
	}
	if ($self->{SET_OP} eq '-') {
	    delete $defaultParms->{$attr};
	} else {
	    $defaultParms->{$attr} = $value;
	}
    }
}
sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    return "flags $self->{SET_OP} (".join(' ', map {$_->toString} @{$self->{BINDINGS}}).")";
}

package W3C::Rdf::AlgaeCompileTree::FwRule;
use W3C::Util::ArrayUtils qw(&DecorateArray);
use W3C::Rdf::Atoms qw($ATTRIB_GroundFact);
my $FwRuleAtoms = {};
sub new {
    my ($proto, $body, $head, @actionParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $FwRuleAtoms->{$class}{$body}{$head};
    if (!$self) {
	$self = $class->SUPER::new(@actionParms);
	$self->{BODY} = $body;
	$self->{HEAD} = $head;
	$FwRuleAtoms->{$class}{$body}{$head} = $self;
    }
    my $atoms = $self->{ALGAE2}{-atomDictionary};
    $self->{RULE_RESULT_SET} = new W3C::Rdf::ChildResultSet($self->{ALGAE2}->getResultSet, -atomDictionary => $atoms);
    $self->{ATTRIBUTION} = new W3C::Rdf::AttributionList();
    $self->{ATTRIBUTION}->ensureDirectAttribution($self->{ALGAE2}->getSourceAttribution);
    return $self;
}
sub delayedEvaluate {
    my ($self, $resultSet) = @_;
    $self->{BODY}->getDB->addForwardChainingRule($self, $self->{BODY}->getDB, $self->{RULE_RESULT_SET}, $self->{ATTRIBUTION});
}
sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    my @a;
    if (!$flags{-brief}) {
	@a = $self->{ATTRIBUTION}->toString(%flags);
    }
    if (wantarray) {
	return $flags{-outputMode} eq 'n3' ? 
	    (' a :FwRule ; ', 
	     &DecorateArray(' :repr {{ ', '         ', '', ' }', $self->{BODY}->toString(%flags)), 
	     '         => ', 
	     &DecorateArray('        { ', '         ', '', ' }} ;', $self->{HEAD}->toString(%flags)), 
	     &DecorateArray(' :accordingTo ', '              ', '', '', @a)) : 
	     &DecorateArray('fwrule ', '       ', '', '', 
			    ($self->{BODY}->toString(%flags), 
			     $self->{HEAD}->toString(%flags), 
			     &DecorateArray('   -->', '      ', '', '', @a)));
    } else {
	my $aStr = @a ? ' '.join (' ', @a) : '';
	return 'fwrule '.$self->{BODY}->toString(%flags).' '.
	    $self->{HEAD}->toString(%flags).$aStr;
    }
}
sub getBody {$_[0]->{BODY}}
sub getHead {$_[0]->{HEAD}}

package W3C::Rdf::AlgaeCompileTree::Clear;
sub new {
    my ($proto, @actionParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@actionParms);
    return $self;
}
sub immediateEvaluate {
    my ($self, $resultSet) = @_;
    $resultSet->clear(); # = new W3C::Rdf::ResultSet(-atomDictionary => $self->{-atomDictionary});
    my $labels = $self->{ALGAE2}->getLabels;
    splice (@$labels, 0, @$labels);
    #my $collects = $self->{ALGAE2}->getCollects;
    #splice (@$collects, 0, @$collects);
    $self->{ALGAE2}{-namespaceHandler}->clear;
}
sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    return 'clear \'('.join (' ', map {$_->toString} @{$self->{NODES}}).')';
}

package W3C::Rdf::AlgaeCompileTree::Collect;
sub new {
    my ($proto, $nodes, @actionParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@actionParms);
    $self->{NODES} = $nodes;
    return $self;
}
sub immediateEvaluate {
    my ($self, $resultSet) = @_;
    foreach my $atom (@{$self->{NODES}}) {
	$atom->addSelect();
    }
#    my $collects = $self->{ALGAE2}->getCollects;
#    push (@$collects, @{$self->{NODES}});
}
sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    my $nodeStr = join (' ', map {$_->toString} @{$self->{NODES}});
    return "collect ($nodeStr)";
}

package W3C::Rdf::AlgaeCompileTree::CollectStar;
sub new {
    my ($proto, @actionParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@actionParms);
    return $self;
}
sub immediateEvaluate {
    my ($self, $resultSet) = @_;
    $self->{RESULT_SET} = $resultSet;
}
sub expandCollects {
    my ($self) = @_;
    my @symbols = $self->{RESULT_SET}->getSymbols();
    for (my $i = 0; $i < @symbols; $i++) {
	$self->{ALGAE2}->addLabel($symbols[$i], $i);
    }
}
sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    return "collect (*)";
}

package W3C::Rdf::AlgaeCompileTree::Host;
sub new {
    my ($proto, $nodes, @actionParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@actionParms);
    $self->{NODES} = $nodes;
    return $self;
}
sub immediateEvaluate {
    my ($self, $resultSet) = @_;
#    $resultSet->skipUsedColumns;
    my $hostables = $self->{ALGAE2}->getHostSymbols;
    foreach my $node (@{$self->{NODES}}) {
	$hostables->{$node->toString} = $node;
    }
}
# Here's sort of what you'd need to do the dalayedEval version of Host.
sub delayedEvaluateNUKE {
    my ($self, $resultSet) = @_;
    foreach my $atom (@{$self->{NODES}}) {
	my $newId = $self->{ALGAE2}{-atomDictionary}->getUri($self->{ALGAE2}->getNextHostID);
	$self->{DB}->changeId(undef, $atom, $newId);
	$resultSet->changeId($atom, $newId);
    }
}
sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    return 'collect \'('.join (' ', map {$_->toString} @{$self->{NODES}}).')';
}

package W3C::Rdf::AlgaeCompileTree::Slurp;
sub new {
    my ($proto, $src, $parms, $optDbSpec, $loader, @actionParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@actionParms);
    if (!$loader) {
	&throw(new W3C::Util::Exception(-message => 'slurp needs an loader.'));
    }
    $self->{SOURCE} = $src;
    $self->{PARMS} = $parms;
    $self->{DB_SPEC} = $optDbSpec;
    $self->{LOADER} = $loader;
    return $self;
}
sub immediateEvaluate {
    my ($self, $resultSet) = @_;
    my $startState = undef;
    my $parms = $self->{PARMS};
    my $rdfApp = $self->{LOADER};
    my $inputSource = &W3C::Rdf::RdfApp::getInputSource($self->{SOURCE}->getUrl, $parms);
    if ($parms->{-printData}) {
	my $handle = $parms->{-printData};
	print $handle $inputSource->getByteStream;
    }
    my $origStartState = $rdfApp->{ARGS}{'-startState'};
    if (defined $startState) {
	# each startState is good for one parsing
	$rdfApp->{ARGS}{'-startState'} = $startState;
	$startState = undef;
    }
    my $lang = $parms->{-inputLang} ? $parms->{-inputLang}->getString : undef;
    my $origDB = $rdfApp->getRdfDB();
    if ($self->{DB_SPEC} && $origDB != $self->{DB_SPEC}) {
	$rdfApp->setRdfDB($self->{DB_SPEC});
    }
    $rdfApp->parseOne($inputSource, $lang);
    if ($self->{DB_SPEC} && $origDB != $self->{DB_SPEC}) {
	$rdfApp->setRdfDB($origDB);
    }
    $rdfApp->{ARGS}{'-startState'} = $origStartState;

}
sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    my $dbStr = ''; #!!! $self->{DB_SPEC} ? $self->{DB_SPEC}->toString.' ' : '';
    my $sourceStr = $self->{SOURCE}->toString;
    my $parmsStr = ''; #!!! ($self->{PARMS} ? join (' ', %{$self->{PARMS}}) : '');
    return "slurp $sourceStr $dbStr ($parmsStr)";
}

# TriplesActions
package W3C::Rdf::AlgaeCompileTree::TriplesAction;
use W3C::Util::ArrayUtils qw(&DecorateArray);
my $TriplesActions = {};
sub new {
    my ($proto, $decls, $optDbSpec, $defaultDB, @actionParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $TriplesActions->{$class}{$decls->toString}{$optDbSpec}{$defaultDB};
    if (!$self) {
	$self = $class->SUPER::new(@actionParms);
	$self->{DECLS} = $decls;
	$self->{DB_SPEC} = $optDbSpec;
	$self->{DEF_DB} = $defaultDB;
	$TriplesActions->{$class}{$decls->toString}{$optDbSpec}{$defaultDB} = $self;
    }
    return $self;
}
sub getDB {
    my ($self) = @_;
    return $self->{DB_SPEC} ? $self->{DB_SPEC} : $self->{DEF_DB};
}
sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    my $dbStr = $self->{DB_SPEC} && $self->{DB_SPEC}{-name} ? 
	$self->{DB_SPEC}{-name}.' ' :
	'';
    my $leadStr = $flags{-outputMode} eq 'n3' ? "" : $self->getActionName."$dbStr(";
    my @declStrs = $self->{DECLS}->toString(%flags);
    if (wantarray) {
	my $indent = ' ' x (length($leadStr));
	return &DecorateArray($leadStr, $indent, '', $flags{-outputMode} eq 'n3' ? '' : ')', @declStrs);
    } else {
	my $declStr = join (' ', @declStrs);
	return $leadStr.$declStr.')';
    }
#    return $self->getActionName." '($dbStr\n".join ("\n", map {' '.$_->toString} @{$self->{DECLS}}).')';
}
sub getCNF {$_[0]->{DECLS}->toCNF}
package W3C::Rdf::AlgaeCompileTree::Ask;
sub getActionName {'ask'}
sub delayedEvaluate {
    my ($self, $resultSet) = @_;
#    $resultSet->atLeastOneRow;
    $self->getDB()->dbEvaluateQTerm($resultSet, $self->{DECLS}, $self->getDB());
    $self->{ALGAE2}->dumpStuff('Query', $self->{ALGAE2}, $self->{DECLS});
    $self->{ALGAE2}->dumpStuff('QueryGraph', $self->{ALGAE2}, $self->{DECLS});
}
package W3C::Rdf::AlgaeCompileTree::Assert;
sub getActionName {'assert'}
sub delayedEvaluate {
    my ($self, $resultSet, $visited, $trigger, $attribution) = @_;if (@_ < 4) {&throw}
#    $resultSet->atLeastOneRow;
    $self->getDB()->dbEvaluateAssertion($resultSet, $self->{DECLS}, $self->getDB(), $visited, $trigger, $attribution);
}


package W3C::Rdf::AlgaeCompileTree::Thing;
use W3C::Util::Exception;
sub new {
    my ($proto, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    $exprParms[0] || &throw(new W3C::Util::Exception(-message => "Bad exprParms (@exprParms) for $class baseclass Thing"));
    my $self = {ALGAE2 => $exprParms[0]->YYData->{ALGAE2}, 
		PARSER => $exprParms[0], 
		LASTPOS => $exprParms[0]->YYData->{my_LASTPOS}};
    bless ($self, $class);
    return $self;
}
sub NPnew {
    my ($proto, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    $exprParms[0] || &throw(new W3C::Util::Exception(-message => "Bad exprParms (@exprParms) for $class baseclass Thing"));
    my $self = {ALGAE2 => $exprParms[0], PARSER => undef, LASTPOS => undef};
    bless ($self, $class);
    return $self;
}

sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    return "!!! $self no toString defined";
}

sub val {
    my ($self, $triple, $premise, $row) = @_;
    return "!!! $self no val defined";
}

package W3C::Rdf::AlgaeCompileTree::Decl;
use W3C::Util::Exception;
use W3C::Rdf::Atoms qw($Value_NULL);
use vars qw($DISJUNCTION $NEGATION $OUTER);
($DISJUNCTION, $NEGATION, $OUTER) = ($W3C::Rdf::AlgaeCompileTree::DISJUNCTION, $W3C::Rdf::AlgaeCompileTree::NEGATION, $W3C::Rdf::AlgaeCompileTree::OUTER);
sub new {
    my ($proto, $parts, $constraints, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@exprParms);
    $self->{PARTS} = $parts;
    $self->{CONSTRAINTS} = $constraints;
    return $self;
}
sub NPnew {
    my ($proto, $parts, $constraints, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::NPnew(@exprParms);
    $self->{PARTS} = $parts;
    $self->{CONSTRAINTS} = $constraints;
    return $self;
}
sub setLastConstraints {$_[0]->{CONSTRAINTS} = $_[1]}
sub val999 { # @@@ could be used if a triple meant something in a constraint
    my ($self, $triple, $premise, $row) = @_;
    return $self->{PARTS}; # !!!
}
sub arity {3}
sub getSlot {$_[0]->{PARTS}[$_[1]]}
sub evaluateAssertion {
    my ($self, $resultSet, $db, $visited, $trigger, $attribution) = @_;
    my @insertions = $self->assertedTriples($resultSet, $db, $visited, $trigger, $attribution);
    foreach my $insertion (@insertions) {
	my ($triple, $reified) = @$insertion;
	push (@$reified, $db->addTriple($triple));
	my $statement = $db->applyRules([$triple], $triple, $visited, $attribution);
	1;#my $statement = $db->applyRules($reified, $triple);
    }
}
sub assertedTriples {
    my ($self, $resultSet, $db, $visited, $trigger, $attribution) = @_; # !!! nuke $db if possible
    my @insertions;
    my $usedState = $resultSet->{UNUSED};
    if ($self->_allStatic()) {
	push (@insertions, $self->_foo(undef, $db, $visited, $attribution));
    } elsif (!$resultSet->isFresh()) {
	for (my $e = $resultSet->elements; $e->hasMoreElements;) {
	    my $row = $e->nextElement;
	    $resultSet->markUsed;
	    # print $row->toString,"\n";
	    my $substAttrib = $self->{ALGAE2}{-atomDictionary}->getInferenceAttribution($attribution, [$row->getProofs()], {$row->getBindings()});
	    push (@insertions, $self->_foo($row, $db, $visited, $substAttrib));
	}
    }
    $resultSet->{UNUSED} = $usedState;
    return @insertions;
}
sub _allStatic {
    my ($self) = @_;
    foreach my $part (@{$self->{PARTS}}) {
	if (!$part->isa('W3C::Rdf::AlgaeCompileTree::StaticPOS')) {
	    return 0;
	}
    }
    return 1;
}
sub _foo {
    my ($self, $row, $db, $visited, $attribution) = @_;
    my @statementParms = ();
    for (my $iParm = 0; $iParm < 4; $iParm++) {
	my $atom = $iParm < $self->arity ? 
	    $self->getSlot($iParm)->lookfor($row, $db, $self->{ALGAE2}{-atomDictionary}) : 
	    undef;
	if (ref $atom eq 'SCALAR') {
	    $atom = $row->get($$atom);
	}
	push (@statementParms, $atom);
    }
    # at this time, uri2Attribution is not defined in any db.
    #$self->{ALGAE2}->getSourceAttribution; # !!! use ATTRIB
    $statementParms[4] = $self->arity > 4 ? $db->uri2Attribution($self->[4]) : $attribution;

    # Ideally, we would create a statement and add it to the database.
    # The problem is that the tests need to have a statement to work on.
    # But those tests my keep the statement from being created. So we
    # create the statement, try the tests, and then delete it if the tests
    # indicated it should not be created.
    # Following is a bit of code that recreates the following statement:
    # my ($reified, $triple) = $db->createStatement(@statementParms);
    my $triple;
    my $pendingException = undef;
    eval {
	my ($p, $s, $o, $r, $a) = @statementParms;
	eval {
	    $triple = $self->{ALGAE2}{-atomDictionary}->getStatement($p, $s, $o, $r, $a);
	}; if ($@) {&throw();}
	# $triple = new W3C::Rdf::MappedTriple($db, @statementParms[0..2,4]);
    }; if ($@) {if (my $ex = &catch('W3C::Rdf::BadlyFormedTripleException')) {
	$triple = $ex->getTriple;
	$pendingException = $ex;
    } elsif (my $ex = &catch('W3C::Util::Exception')) {
	&throw($ex);
    } else {
	die $@;
    }}

    if (exists $visited->{$triple}{$attribution}) { # maybe if $why ? {$why->[0]}{$why->[1]}
	next;
    }
#    if ($db->triplesMatching(undef, [[@statementParms[0..2]]])) {
#	1;#next;
#    }
    $visited->{$triple}{$attribution} = undef; # {$why->[0]}{$why->[1]}
    my $reified = $db->reify($triple, $statementParms[3]);

    # evalulate constraints
    if ($self->{CONSTRAINTS}) {
	foreach my $constraint (@{$self->{CONSTRAINTS}}) {
	    my $res = $constraint->val($triple, $self, $row);
	    if (!$res || ($res->isa('W3C::Rdf::String') && !$res->getString)) {
		return;
	    }
	}
    }

    if ($pendingException) {
	&throw($pendingException);
    }

    return [$triple, $reified];
}
sub evaluateQTerm {
    my ($self, $resultSet, $db, $modifier) = @_;
    for (my $e = $resultSet->elements; $e->hasMoreElements;) {
	my $row = $e->nextElement;
	$resultSet->markUsed;
	my @tmp = $self->_matchDecl($row, $db);

	# evalulate constraints
	if ($self->{CONSTRAINTS}) {
	    foreach my $constraint (@{$self->{CONSTRAINTS}}) {
		for (my $iT = 0; $iT < @tmp; $iT++) {
		    my $passed = 1;
		    eval {
			my $res = $constraint->val($tmp[$iT], $self, $row);
			if (!$res || ($res->isa('W3C::Rdf::String') && !$res->getString)) {
			    $passed = 0;
			}
		    }; if ($@) {if (my $ex = &catch('W3C::Rdf::SafeEvaluationException')) {
			$passed = 0;
		    } else {
			&throw();
		    }}

		    if (!$passed) {
			splice (@tmp, $iT, 1);
			--$iT;
		    }
		}
	    }
	}

	# absorb the results into the data set
	# comments assume changeing
	# [1 1] [s1 s2]
	# [2 4] [s4 s5]
	# [3 9] [s7 s8]
	# to
	# [1 1 1] [s1 s2 s3]
	# [2 4 8] [s4 s5 s6]
	# [3 9 27] [s7 s8 s9]
	if (@tmp) {
	    foreach my $triple (@tmp) {
		# build a new array of results (and supporting statements) in a new array
		# This could be done with an array splice, but it's less efficient and less portable to C/Java.

		# Next results start with current row ([2, 4]).
		#  and current statements ([s4 s5]).
		my $newRow = $row->duplicate;
		# Add this statement (s6) to the proofs ==> ([s4 s5 s6]).
		if ($self->{ALGAE2}{-flags}->{-lowLevelStatements} && @{$self->{ALGAE2}{-flags}->{-lowLevelStatements}}) {
		    my $statements = shift @{$self->{ALGAE2}{-flags}->{-lowLevelStatements}};
		    $newRow->addProofs($statements);
		} else {
		    $newRow->addProofs([$triple]);
		}
		# Take results (8) from this statement (s6) => ([2 4 8]).
		for (my $i = 0; $i < $self->arity; $i++) {
		    $self->getSlot($i)->absorb($triple->getPosition($i), $newRow);
		}
	    }
	    $row->eliminate;
	} elsif ($modifier & $OUTER) {
	    # keep row
	    # assign NULLs to otherwise unassigned variable slots
	    for (my $i = 0; $i < $self->arity; $i++) {
		my $part = $self->getSlot($i);
		if ($part->isa('W3C::Rdf::AlgaeCompileTree::Var') && !defined $row->get($part->varIndex)) {
		    $part->absorb($Value_NULL, $row);
		}
	    }
	} else {
	    $row->eliminate;
	}
    }
}

sub _matchDecl {
    my ($self, $row, $db) = @_;
    my $lookfor = [];
    if (!$self->{ALGAE2}{-flags}->{-lowLevelStatements}) {
	$self->{ALGAE2}{-flags} = {%{$self->{ALGAE2}{-flags}}, -lowLevelStatements => []};
    }
    for (my $i = 0; $i < $self->arity; $i++) {
	my $atom = $self->getSlot($i)->lookfor($row, $self->{-atomDictionary});
	if ($atom == $Value_NULL) {
	    return ();
	}
	if (ref $atom eq 'SCALAR') {
	    $atom = $row->get($$atom);
	}
	push (@$lookfor, $atom);
    }
    my @ret = $db->triplesMatching(undef, [$lookfor], $self->{ALGAE2}{-flags});
    if (@ret < $self->{-minimumRowSolution}) {
	my $flags = {-minimumRowSolution => $self->{-minimumRowSolution}};
	$self->minimumRowSolutionMethod($self, $flags, \@ret, $row, $lookfor);
    }

    # trim for cases like (dc:title dc:title "title")
    for (my $i = 0; $i < $self->arity - 1; $i++) {
	for (my $j = $i+1; $j < $self->arity; $j++) {
	    if ($self->getSlot($i)->matchingIntangible($self->getSlot($j))) {
		for (my $iT = 0; $iT < @ret; ++$iT) {
		    if ($ret[$iT]->getPosition($i) != $ret[$iT]->getPosition($j)) {
			splice (@ret, $iT, 1);
			--$iT;
		    }
		}
	    }
	}
    }

    return @ret;
}

sub toCNF {$_[0]}

# @@@ not operational right now
sub minimumRowSolutionMethod {
    my ($self, $term, $flags, $initialSolutions, $row, $lookfor, $db, $view, $querySets) = @_;
    # note - will want to vary constants but not bound variabels for tab completion.
    print STDERR 'failed: '.$term->getSlot(0)->getExpr->toString."\n";

    my ($contants , $boundVars, $unboundVars) = (0, 0, 0);
    for (my $i = 0; $i < $term->arity; $i++) {
	if ($term->getSlot($i)->isa('W3C::Rdf::Algae::ConstantQueryPiece')) {
	    $contants++;
	} elsif ($term->getSlot($i)->isa('W3C::Rdf::Algae::VariableQueryPiece')) {
	    if ($lookfor->[$i]) {
		$boundVars++;
	    } else {
		$unboundVars++;
	    }
	}
    }

    # check variations on new variables
    for (my $unbind = 1; $unbind < $boundVars + $contants; $unbind++) {
	# try substituting $unbind contants
	for (my $i = 0; $i < $term->arity; $i++) {
	    if ($term->getSlot($i)->isa('W3C::Rdf::Algae::ConstantQueryPiece')) {
		my $tmp = $lookfor->[$i];
		$lookfor->[$i] = undef;
		@$initialSolutions = $db->triplesMatching($view, [$lookfor], $flags);
		for (my $iWorky = 0; $iWorky < @$initialSolutions; $iWorky++) {
		    my ($try, $place) = 
			$i == 0 ? ($initialSolutions->[$iWorky]->getPredicate, 'predicate') : $i == 1 ? ($initialSolutions->[$iWorky]->getSubject, 'subject') : $i == 2 ? ($initialSolutions->[$iWorky]->getObject, 'object') : (undef, undef);
		    $try = $try->toString;
		    print STDERR "trying $place $try\n";
		}
	    }
	}
    }
    if ($flags && @$initialSolutions < $flags->{-minimumRowSolution}) {
	my @bindings;
	foreach my $term ($querySets->getTerms) {
	    for (my $j = 0; $j < $term->arity; $j++) {
		my $qp = $term->[$j];
		if ($qp->isa('W3C::Rdf::Algae::VariableQueryPiece')) {
		    my $column = $qp->varIndex;
		    my $sym = $qp->symbol;
		    my $val = $row->[$column];
		    if (defined $val) {
			$val = $val->toString;
		    } else {
			last;
		    }
		    $bindings[$column] = "?$sym $val";
		}
	    }
	}
	print STDERR "dropping row:\n    ".join("\n    ", @bindings)."\n";
    }
}

sub toGraph {
    my ($self, $or, $not, $opt) = @_;
    my @pParts;
    if ($or) {push @pParts, '||'}
    if ($not) {push @pParts, '!'}
    if ($opt) {push @pParts, '~'}
    push (@pParts, $self->{PARTS}[0]->toString(undef, undef));
    my $s = $self->{PARTS}[1]->toString(undef, undef);
    $s =~ s/(\\(?!n)|\"|<|>)/\\$1/gx;
    my $o = $self->{PARTS}[2]->toString(undef, undef);
    $o =~ s/(\\(?!n)|\"|<|>)/\\$1/gx;
    if ($self->{CONSTRAINTS}) {
	push (@pParts, join ("\\n", map {$_->toString(undef, undef)} @{$self->{CONSTRAINTS}}));
    }
    map {s/(\\(?!n)|\"|<|>)/\\$1/gx} @pParts;
    return "\"$s\" -> \"$o\" [label=\"".join (' ', @pParts)."\"];\n";
}

sub toDot {
    my ($self, %flags) = &main::_defaultPrec(@_);
    my @parts;
    foreach my $i (1, 0, 2) {
	push (@parts, $self->{PARTS}[$i]->toString(%flags));
    }
    if ($self->{CONSTRAINTS}) {
	push (@parts, join ("\\n", map {$_->toString(%flags)} @{$self->{CONSTRAINTS}}));
    }
    map {s/(\\(?!n)|\"|<|>)/\\$1/gx} @parts;
    return "\"$self\" [shape=record label=\"".join ('|', @parts)."\"];\n";
}

sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    my $ret = join (' ', map {$_->toString(%flags)} 
		    ($self->{PARTS}[1], 
		     $self->{PARTS}[0], 
		     $self->{PARTS}[2]));
    if ($self->{CONSTRAINTS}) {
	$ret .= ' '.join('', map {'{'.$_->toString(%flags, -parentPrec => $MAX_PREC).'}'} @{$self->{CONSTRAINTS}});
    }
    return $ret;
}

package W3C::Rdf::AlgaeCompileTree::UnaryDecl;
sub new {
    my ($proto, $decl, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@exprParms);
    $self->{DECL} = $decl;
    return $self;
}
sub getDecl {$_[0]->{DECL}}
sub toDot {
    my ($self, %flags) = &main::_defaultPrec(@_);
    my $opStr = $self->opStr;
    return "\"$self\" [label=\"$opStr\"];
\"$self\" -> \"$self->{DECL}\"
".$self->{DECL}->toDot(%flags);
}
sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    my $opStr = $self->opStr;
    return $opStr.$self->{DECL}->toString(%flags);
}


package W3C::Rdf::AlgaeCompileTree::Negation;
use W3C::Util::Exception;
use W3C::Rdf::Atoms qw($Value_NULL);
sub evaluateQTerm {
    my ($self, $resultSet, $db, $modifier) = @_;

    for (my $e = $resultSet->elements; $e->hasMoreElements;) {
	my $row = $e->nextElement;
	my $miniResultSet = $row->makeResultSet;
	$self->{DECL}->evaluateQTerm($miniResultSet, $db, $modifier);
	my $miniE = $miniResultSet->elements;
	if ($miniE->hasMoreElements) {
	    # if there are unbound variables, we
	    #   push (@$nextResults, [@$row]);
	    #   push (@$nextStatements, [@$statements]);
	    # and make sure we don't get those bindings later...
	    #   for (my $i = 0; $i < $self->arity; $i++) {
	    #       $self->getSlot($i)->absorb($triple->getPosition($i), $nextResults, NOT!);
	    #   }
	    $row->eliminate;
	}
    }
}
sub toGraph {
    my ($self, $or, $not, $opt) = @_;
    $self->{DECL}->toGraph($or, !$not, $opt);
}
sub opStr {'!'}

package W3C::Rdf::AlgaeCompileTree::Option;
use W3C::Util::Exception;
use W3C::Rdf::Atoms qw($Value_NULL);
use vars qw($DISJUNCTION $NEGATION $OUTER);
($DISJUNCTION, $NEGATION, $OUTER) = ($W3C::Rdf::AlgaeCompileTree::DISJUNCTION, $W3C::Rdf::AlgaeCompileTree::NEGATION, $W3C::Rdf::AlgaeCompileTree::OUTER);
sub evaluateQTerm {
    my ($self, $resultSet, $db, $modifier) = @_;

    for (my $e = $resultSet->elements; $e->hasMoreElements;) {
	my $row = $e->nextElement;
	my $miniResultSet = $row->makeResultSet;
	$self->{DECL}->evaluateQTerm($miniResultSet, $db, $modifier | $OUTER);
	my $empty = 1;
	for (my $miniE = $miniResultSet->elements; $miniE->hasMoreElements;) {
	    $empty = 0;
	    my $miniRow = $miniE->nextElement;
	    my $newRow = $row->duplicate;
	    $newRow->assumeNewBindings($miniRow);
	}
	if (!$empty) {
	    $row->eliminate;
	}
    }
}
sub toGraph {
    my ($self, $or, $not, $opt) = @_;
    $self->{DECL}->toGraph($or, $not, 1);
}
sub opStr {'~'}

package W3C::Rdf::AlgaeCompileTree::BinaryDecl;
use W3C::Util::Exception;
use W3C::Rdf::Atoms qw($Value_NULL);
sub new {
    my ($proto, $left, $right, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@exprParms);
    $self->{LEFT} = $left;
    $self->{RIGHT} = $right;
    return $self;
}
sub getLeft {$_[0]->{LEFT}}
sub getRight {$_[0]->{RIGHT}}
sub getDecls {($_[0]->{LEFT}, $_[0]->{RIGHT})}

sub toDot {
    my ($self, %flags) = &main::_defaultPrec(@_);
    my $opStr = $self->opStr;
    return "\"$self\" [label=\"$opStr\"];
\"$self\" -> \"$self->{LEFT}\"  [label=\"l\"];
\"$self\" -> \"$self->{RIGHT}\"  [label=\"r\"];
".$self->{LEFT}->toDot(%flags)."
".$self->{RIGHT}->toDot(%flags);
}
sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    my $opStr = $self->opStr(%flags);
    if (wantarray) {
	return $self->{LEFT}->toString(%flags)." $opStr ", $self->{RIGHT}->toString(%flags);
    } else {
	return join (" $opStr ", map {$_->toString(%flags)} ($self->{LEFT}, $self->{RIGHT}));
    }
}

package W3C::Rdf::AlgaeCompileTree::Conjunction;
use W3C::Util::Exception;
use W3C::Rdf::Atoms qw($Value_NULL);
sub setLastConstraints {$_[0]->{RIGHT}->setLastConstraints($_[1])}
sub evaluateAssertion {
    my ($self, $resultSet, $db, $visited, $trigger, $attribution) = @_;
    $self->{LEFT}->evaluateAssertion($resultSet, $db, $visited, $trigger, $attribution);
    $self->{RIGHT}->evaluateAssertion($resultSet, $db, $visited, $trigger, $attribution);
}
sub assertedTriples {
    my ($self, $resultSet, $db, $visited, $trigger, $attribution) = @_; # !!! nuke $db if possible
    return ($self->{LEFT}->assertedTriples($resultSet, $db, $visited, $trigger, $attribution), 
	    $self->{RIGHT}->assertedTriples($resultSet, $db, $visited, $trigger, $attribution));
}
sub evaluateQTerm {
    my ($self, $resultSet, $db, $modifier) = @_;
    foreach my $decl ($self->{LEFT}, $self->{RIGHT}) {
	eval {
	    $decl->evaluateQTerm($resultSet, $db, $modifier);
	}; if ($@) {
	    my $ex;
	    if ($ex = &catch('W3C::Rdf::RdfDBException')) {
		$self->{-errorHandler}->warning($ex);
	    } elsif ($ex = &catch('W3C::Util::Exception')) {
#		$self->{-errorHandler}->warning($ex);
		&throw($ex);
	    } else {
		&throw();
		$self->{-errorHandler}->error(new W3C::Util::PerlException);
	    }
	}
    }
}
sub toCNF {
    my ($self) = @_;
    return ($self->{LEFT}->toCNF, $self->{RIGHT}->toCNF);
}
sub toGraph {
    my ($self, $or, $not, $opt) = @_;
    $self->{LEFT}->toGraph($or, $not, $opt)."\n".$self->{RIGHT}->toGraph($or, $not, $opt)."\n";
}
sub opStr {
    my ($self, %flags) = @_;
    return $flags{-outputMode} eq 'n3' ? '.' : '&&';
}

package W3C::Rdf::AlgaeCompileTree::ShortcutDisjunction;
use W3C::Util::Exception;
use W3C::Rdf::Atoms qw($Value_NULL);
sub evaluateQTerm {
    my ($self, $resultSet, $db, $modifier) = @_;

    for (my $e = $resultSet->elements; $e->hasMoreElements;) {
	my $row = $e->nextElement;
	my $empty = 1;
	foreach my $term ($self->{LEFT}, $self->{RIGHT}) {
	    my $miniResultSet = $row->makeResultSet;
	    $term->evaluateQTerm($miniResultSet, $db, $modifier);
	    for (my $miniE = $miniResultSet->elements; $miniE->hasMoreElements;) {
		$empty = 0;
		my $miniRow = $miniE->nextElement;
		my $newRow = $row->duplicate;
		$newRow->assumeNewBindings($miniRow);
	    }
	    last if (!$empty);
	}
	$row->eliminate;
    }
}
sub toGraph {
    my ($self, $or, $not, $opt) = @_;
    $self->{LEFT}->toGraph(1, $not, $opt)."\n".$self->{RIGHT}->toGraph(1, $not, $opt)."\n";
}
sub opStr {'||'}

package W3C::Rdf::AlgaeCompileTree::UnionDisjunction;
use W3C::Util::Exception;
use W3C::Rdf::Atoms qw($Value_NULL);
sub evaluateQTerm {
    my ($self, $resultSet, $db, $modifier) = @_;

    for (my $e = $resultSet->elements; $e->hasMoreElements;) {
	my $row = $e->nextElement;
	foreach my $term ($self->{LEFT}, $self->{RIGHT}) {
	    my $miniResultSet = $row->makeResultSet;
	    $term->evaluateQTerm($miniResultSet, $db, $modifier);
	    for (my $miniE = $miniResultSet->elements; $miniE->hasMoreElements;) {
		my $miniRow = $miniE->nextElement;
		my $newRow = $row->duplicate;
		$newRow->assumeNewBindings($miniRow);
	    }
	}
	$row->eliminate;
    }
}
sub toGraph {
    my ($self, $or, $not, $opt) = @_;
    $self->{LEFT}->toGraph(1, $not, $opt)."\n".$self->{RIGHT}->toGraph(1, $not, $opt)."\n";
}
sub opStr {'||'}

package W3C::Rdf::AlgaeCompileTree::UnionDisjunctionMerge;
use W3C::Util::Exception;
use W3C::Rdf::Atoms qw($Value_NULL);
sub evaluateQTerm {
    my ($self, $resultSet, $db, $modifier) = @_;

    for (my $e = $resultSet->elements; $e->hasMoreElements;) {
	my $row = $e->nextElement;
	my $keepRow = 0;
	foreach my $term ($self->{LEFT}, $self->{RIGHT}) {
	    my $miniResultSet = $row->makeResultSet;
	    $term->evaluateQTerm($miniResultSet, $db, $modifier);
	    for (my $miniE = $miniResultSet->elements; $miniE->hasMoreElements;) {
		my $miniRow = $miniE->nextElement;
		if (my $foundRow = $row->getCousinWithBindings($miniRow)) {
		    $foundRow->assumeNewProofs($miniRow);
		    if ($foundRow == $row) {
			$keepRow = 1;
		    }
		} else {
		    my $newRow = $row->duplicate;
		    $newRow->assumeNewBindings($miniRow);
		}
	    }
	}
	if (!$keepRow) {
	    $row->eliminate;
	}
    }
}
sub toGraph {
    my ($self, $or, $not, $opt) = @_;
    $self->{LEFT}->toGraph(1, $not, $opt)."\n".$self->{RIGHT}->toGraph(1, $not, $opt)."\n";
}
sub opStr {'||'}

package W3C::Rdf::AlgaeCompileTree::POS;
use W3C::Util::Exception;
sub lookfor {&throw(new W3C::Util::MethodNotImplementedException(-object => $_[0], -function => 'W3C::Rdf::AlgaeCompileTree::POS::lookfor'))}
sub absorb ($$$) {&throw(new W3C::Util::MethodNotImplementedException(-object => $_[0], -function => 'W3C::Rdf::AlgaeCompileTree::POS::absorb'))}
sub matchingIntangible ($$) {return undef;}

package W3C::Rdf::AlgaeCompileTree::StaticPOS;
sub lookfor {$_[0]->{INTERNED}}
sub val {$_[0]->{INTERNED}}
sub absorb ($$$) {}
sub matchingIntangible ($$) {return undef;}
sub addSelect {
    my ($self) = @_;
    $self->{ALGAE2}->addLabel($self->toString, $self->{INTERNED});
}
sub symbol ($$) {return $_[0]->{CONST}}

package W3C::Rdf::AlgaeCompileTree::Literal;
sub new {
    my ($proto, $literal, $dt, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@exprParms);
    $self->{CONST} = $literal;
    $self->{DT} = $dt;
    $self->{INTERNED_DT} = $dt ? $self->{ALGAE2}{-atomDictionary}->getUri($dt->getUrl) : undef;
    $self->{INTERNED} = $self->{ALGAE2}{-atomDictionary}->getString($literal, $self->{INTERNED_DT}, 'PLAIN');
    return $self;
}

sub NPnew {
    my ($proto, $literal, $dt, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::NPnew(@exprParms);
    $self->{CONST} = $literal;
    $self->{DT} = $dt;
    $self->{INTERNED_DT} = $dt ? $self->{ALGAE2}{-atomDictionary}->getUri($dt->getUrl) : undef;
    $self->{INTERNED} = $self->{ALGAE2}{-atomDictionary}->getString($literal, $self->{INTERNED_DT}, 'PLAIN');
    return $self;
}

sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    return "\"$self->{CONST}\"";
}

package W3C::Rdf::AlgaeCompileTree::Num;
sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    return $self->{CONST};
}
#sub _binaryOp {
#    my ($self, $r, $op) = @_;
#    return new W3C::Rdf::AlgaeCompileTree::Num(&$op($self->{CONST}, $r->{CONST}), undef, $self->{PARSER});
#}
#sub multiply {$_[0]->_binaryOp($_[1], sub {$_[0] * $_[1]})}
#sub divide {$_[0]->_binaryOp($_[1], sub {$_[0] / $_[1]})}
#sub modulo {$_[0]->_binaryOp($_[1], sub {$_[0] % $_[1]})}
#sub plus {$_[0]->_binaryOp($_[1], sub {$_[0] + $_[1]})}
#sub minus {$_[0]->_binaryOp($_[1], sub {$_[0] - $_[1]})}

package W3C::Rdf::AlgaeCompileTree::Int;
sub new {
    my ($proto, $literal, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    $class->SUPER::new($literal, new W3C::Rdf::AlgaeCompileTree::Url('http://www.w3.org/2001/XMLSchema#int',  @exprParms), @exprParms);
}

package W3C::Rdf::AlgaeCompileTree::Float;
sub new {
    my ($proto, $literal, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    $class->SUPER::new($literal, new W3C::Rdf::AlgaeCompileTree::Url('http://www.w3.org/2001/XMLSchema#float', @exprParms), @exprParms);
}

package W3C::Rdf::AlgaeCompileTree::UrlBase;
sub new {
    my ($proto, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@exprParms);
    $self->{CONST} = '--- improper constructor ---';
    return $self;
}
sub NPnew {
    my ($proto, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::NPnew(@exprParms);
    $self->{CONST} = '--- improper constructor ---';
    return $self;
}
sub _setUrl {
    my ($self, $url) = @_;
    my $base = $self->{ALGAE2}->getSourceAttribution()->getSource()->getUri();
    my $uriOb = new_abs URI($url, $base);
    $self->{CONST} = $uriOb->canonical();
    $self->{INTERNED} = $self->{ALGAE2}{-atomDictionary}->getUri($uriOb->canonical());
}
sub getUrl {$_[0]->{CONST}}
sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
#    my ($ns, $prefix, $localName) = $self->{ALGAE2}{-namespaceHandler}->unmapNamespace($self->{CONST});
    my ($ns, $prefix, $localName);
    if (my $nsh = $flags{-namespaceHandler}) {
	($ns, $prefix, $localName) = $nsh->unmapNamespace($self->{CONST});
    }
    return defined $ns ? $prefix eq '' ? $localName : "${prefix}:$localName" : "<$self->{CONST}>";
}
sub interned {$_[0]->{INTERNED}}

package W3C::Rdf::AlgaeCompileTree::Url;
sub new {
    my ($proto, $url, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@exprParms);
    $self->_setUrl($url);
    return $self;
}

sub NPnew {
    my ($proto, $url, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::NPnew(@exprParms);
    $self->_setUrl($url);
    return $self;
}

package W3C::Rdf::AlgaeCompileTree::DelayedQName; use W3C::Util::Exception;
sub new {
    my ($proto, $prefix, $localName, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@exprParms);
    $self->{PREFIX} = $prefix;
    $self->{LOCAL_NAME} = $localName;
    return $self;
}
sub resolveNS {
    my ($self) = @_;
    eval {
	my ($ns, $localName, $prefix) = $self->{ALGAE2}{-namespaceHandler}->mapNamespace("$self->{PREFIX}:$self->{LOCAL_NAME}");
	$self->_setUrl("$ns$localName");
    }; if ($@) {if (my $ex = &catch('W3C::Util::UnknownNamespaceException')) {
	my $newEx = new W3C::Util::CachedContextException(-str => $self->{PARSER}->YYData->{INPUT}, 
							  -pos => $self->{LASTPOS}, 
							  -contextID => $self->{PARSER}, 
							  -chainedException => $ex);
	$newEx->assumeStackTrace($ex);
	&throw($newEx);
    } else {
	&throw();
    }}
}

package W3C::Rdf::AlgaeCompileTree::QName;
sub new {
    my ($proto, $prefix, $localName, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new($prefix, $localName, @exprParms);
    $self->resolveNS();
    return $self;
}

package W3C::Rdf::AlgaeCompileTree::BNode;
sub new {
    my ($proto, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@exprParms);
    $self->{INTERNED} = $self->{ALGAE2}{-atomDictionary}->createBNode($self->{ALGAE2}->getSourceAttribution);
    return $self;
}

sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    my $str = $self->{INTERNED}->toString;
    return "[$str]";
}

package W3C::Rdf::AlgaeCompileTree::Var;
use W3C::Util::Exception;
sub new {
    my ($proto, $name, $create, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@exprParms);
    $self->{VARNAME} = $name;
    $self->{NEWSYMBOL} = 0;
    my $resultSet = $self->{ALGAE2}->getResultSet;
    $self->{VARINDEX} = $resultSet->getIndex($name);
    if (!defined $self->{VARINDEX}) {
	if ($create) {
	    $self->{VARINDEX} = $resultSet->bindSymbol($name);
	    $self->{NEWSYMBOL} = 1;
	} else {
	    &throw(new W3C::Util::Exception(-message => "no bindings for $self->{VARNAME}"));
	}
    }
    return $self;
}
sub addSelect {
    my ($self) = @_;
    $self->{ALGAE2}->addLabel($self->{VARNAME}, $self->{VARINDEX});
}
sub varIndex () {return $_[0]->{VARINDEX}}
sub newSymbol () {return $_[0]->{NEWSYMBOL}}

sub symbol ($$) {return $_[0]->{VARNAME}}
sub lookfor ($$) {
    my ($self, $row, $atoms) = @_;
    return \ $self->varIndex;
    return $row->get($self->varIndex);
}

sub absorb ($$$) {
    my ($self, $absorbed, $row) = @_;
    $absorbed = $self->{ALGAE2}->maybeHost($absorbed, $self);
    $row->set($self->varIndex, $absorbed) if ($row); # may be undef for static asserts
    return $absorbed;
}

sub matchingIntangible ($$) {
    my ($self, $against) = @_;
    if ($against->isa('W3C::Rdf::AlgaeCompileTree::Var')) {
	return $self->{VARNAME} eq $against->{VARNAME};
    } else {
	return undef;
    }
}

sub FindOrMake999 {
    my ($name) = @_;
    my $ret = $CalcTree::VARS->{$name};
    if (!$ret) {
	$ret = new Var($name);
	$CalcTree::VARS->{$name} = $ret;
    }
    return $ret;
}

sub val {
    my ($self, $triple, $premise, $row) = @_;
    my $ret = $row->get($self->varIndex);
    if (defined $ret) {
	return $ret;
    }
    for (my $i = 0; $i < $premise->arity; $i++) {
	my $pos = $premise->getSlot($i);
	if ($pos->isa('W3C::Rdf::AlgaeCompileTree::Var') && 
	    $pos->{VARNAME} eq $self->{VARNAME}) {
	    return $triple->getPosition($i);
	}
    }
    return undef;
}
sub assign {
    my ($self, $triple, $premise, $row, $value) = @_;
    $self->absorb($value, $row);
    return $value;
}

sub existential {0}
sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    return '?'.$self->{VARNAME};
}

package W3C::Rdf::AlgaeCompileTree::KeyName;
use W3C::Rdf::Atoms qw($ATTRIB_GroundFact);

sub new {
    my ($proto, $name, $create, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@exprParms);
    $self->{VARNAME} = $name;
    return $self;
}
sub getName {$_[0]->{VARNAME}}
sub val {
    my ($self, $triple, $premise, $row) = @_;
    return &getTripleAttribute($self->{VARNAME}, $triple);
}
sub assign {
    my ($self, $triple, $premise, $row, $value) = @_;
    return &setTripleAttribute($self->{VARNAME}, $triple, $value, $self->{ALGAE2}{-atomDictionary});
}
sub toString {$_[0]->{VARNAME}}

# Access to special variables
sub getTripleAttribute { # static
    my ($varname, $triple) = @_;
    if ($varname eq 'ENCODING') {
	my $o = $triple->getObject;
	return $o->isa('W3C::Rdf::String') ? $o->getEncoding : undef;
    } elsif ($varname eq 'DATATYPE') {
	my $o = $triple->getObject;
	return $o->isa('W3C::Rdf::String') ? $o->getDatatype : undef;
    } elsif ($varname eq 'LANG') {
	my $o = $triple->getObject;
	return $o->isa('W3C::Rdf::String') ? $o->getLang : undef;
    } elsif ($varname eq 'ATTRIB') {
	return $triple->getAttributionList->getFirstUri; # @@@
    } else {
	&throw(new W3C::Util::Exception(-message =>"unknown key word \"$varname\""));
    }
}
sub setTripleAttribute { # static
    my ($varname, $triple, $value, $atomDictionary) = @_;
    if ($varname eq 'ENCODING') {
	$triple->getObject->setEncoding($value);
    } elsif ($varname eq 'DATATYPE') {
	$triple->getObject->setDatatype($value);
    } elsif ($varname eq 'LANG') {
	$triple->getObject->setLang($value);
    } elsif ($varname eq 'ATTRIB') {
	my $auth = undef; # !!!
	$triple->getAttributionList->
	    setDirectAttribution($atomDictionary->
				 getGroundFactAttribution($value, undef, 
							  $auth, undef));
    } else {
	&throw(new W3C::Util::Exception(-message =>"unknown key word \"$varname\""));
    }
    return $value;
}

# ------- CONNECTIVES -------
package W3C::Rdf::AlgaeCompileTree::Constraint;
use W3C::Util::Exception;
sub new {
    my ($proto, $expr, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@exprParms);
    $self->{EXPR} = $expr;
    return $self;
}

sub val {
    my ($self, $triple, $premise, $row) = @_;
    return $self->{EXPR}->val($triple, $premise, $row);
}

sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    return $self->{EXPR}->toString(%flags);
}

sub toSQL {
    my ($self, $algae, $term, $cWheres, $outer) = @_;
    if ($self->{EXPR}) {
	$self->{EXPR}->toSQL($algae, $term, $cWheres, $outer);
    }
}

package W3C::Rdf::AlgaeCompileTree::Binary;
sub new {
    my ($proto, $L, $R, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@exprParms);
    $self->{L} = $L;
    $self->{R} = $R;
    return $self;
}

sub val {
    my ($self, $triple, $premise, $row) = @_;
    return "!!! $self no val defined for $self->{L} ? $self->{R}";
}

sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
    my $opStr = $self->opStr();
    if ($flags{-precTable} && exists $flags{-precTable}{$opStr} && 0) {
	my ($prec, $assoc) = @{$flags{-precTable}{$opStr}};
	my $lStr = $self->{L}->toString(%flags, -parentPrec => $prec);
	my $rStr = $self->{R}->toString(%flags, -parentPrec => $prec);
	if ($prec <= $flags{-parentPrec}) {
	    return "$lStr $opStr $rStr";
	} else {
	    return "($lStr $opStr $rStr)";
	}
    } else {
	my $lStr = $self->{L}->toString();
	my $rStr = $self->{R}->toString();
	return "($lStr $opStr $rStr)";
    }
}

package W3C::Rdf::AlgaeCompileTree::BinaryTest;
sub toSQL {
    my ($self, $algae, $term, $cWheres, $outer) = @_;
    my $l = $self->{L};
    my $r = $self->{R};
    if ($l->isa('W3C::Rdf::AlgaeCompileTree::Var')) {
	if ($r->isa('W3C::Rdf::AlgaeCompileTree::Num')) {
	    my ($firstAlias, $firstField) = $algae->_getFieldForVariable($term, $l);
	    push (@$cWheres, $algae->_addWhereConst($firstAlias, $firstField, $self->opStr(), $r->symbol, $term, $outer));
	} else {
	    &throw(new W3C::Util::ProgramFlowException());
	}
    }
}

package W3C::Rdf::AlgaeCompileTree::Expr;
use W3C::Util::Exception;
use vars qw($XSD_INT $XSD_FLOAT $STR $NUM $URI $BNODE $TYPE_MAP);
($XSD_INT, $XSD_FLOAT) = ('http://www.w3.org/2001/XMLSchema#int', 'http://www.w3.org/2001/XMLSchema#float');
($STR, $NUM, $BNODE, $URI) = (1..4);
$TYPE_MAP = {};
sub internedInt {
    my ($self, $value) = @_;
    # $value = 0 if ($value eq ''); # By default, perl returns '' for (1 == 2) # !!!
    my $dt = $_[0]->{ALGAE2}{-atomDictionary}->getUri($XSD_INT);
    return $_[0]->{ALGAE2}{-atomDictionary}->getString("$value", $dt, 'PLAIN');
}
sub toNum {
    my ($self, $node) = @_;
    if ($node->isa('W3C::Rdf::String')) {
	if (my $dt = $node->getDatatype) {
	    if ($dt->getUri eq $XSD_INT || 
		$dt->getUri eq $XSD_FLOAT) {
		return $node->getString();
	    }
	}
    }
    my $nodeStr = $node->toString();
    &throw(new W3C::Util::Exception(-message => "unable to convert $nodeStr to a NUM"));
}
sub guessType {
    my ($self, $node) = @_;
    if ($node->isa('W3C::Rdf::String')) {
	if (my $dt = $node->getDatatype) {
	    if ($dt->getUri eq $XSD_INT || 
		$dt->getUri eq $XSD_FLOAT) {
		return $NUM;
	    } else {
		return $STR;
	    }
	} elsif ($node->getString =~ m/^\d+(?:\.\d*)?/ || 
		 $node->getString =~ m/^\.\d+/) {
	    return $NUM;
	} else {
	    return $STR;
	}
    } elsif ($node->isa('W3C::Rdf::Uri')) {
	return $URI
    } elsif ($node->isa('W3C::Rdf::BNode')) {
	return $BNODE;
    } else {
	&throw(new W3C::Util::ProgramFlowException());
    }
}
sub cmp {
    my ($self, $l, $r) = @_;
    my ($lt, $rt) = ($self->guessType($l), $self->guessType($r));
    # sort order same as assignments for consts above.
    if ($lt != $rt) {
	return $lt <=> $rt;
    }
    if ($lt == $STR) {
	return $l->getString cmp $r->getString;
    }
    if ($lt == $NUM) {
	return $l->getString <=> $r->getString;
    }
    if ($lt == $BNODE) {
	return $l->getId <=> $r->getId;
    }
    if ($lt == $URI) {
	return $l->getUri cmp $r->getUri;
    }
    &throw(new W3C::Util::ProgramFlowException());
}


sub W3C::Rdf::AlgaeCompileTree::Eq::opStr {'=='}
sub W3C::Rdf::AlgaeCompileTree::Eq::val {$_[0]->internedInt($_[0]->{L}->val(@_[1..3]) == $_[0]->{R}->val(@_[1..3]) ? 1 : 0)}

sub W3C::Rdf::AlgaeCompileTree::Ne::opStr {'!='}
sub W3C::Rdf::AlgaeCompileTree::Ne::val {$_[0]->internedInt($_[0]->{L}->val(@_[1..3]) != $_[0]->{R}->val(@_[1..3]) ? 1 : 0)}

sub W3C::Rdf::AlgaeCompileTree::Glob::opStr {'@'}
sub W3C::Rdf::AlgaeCompileTree::Glob::val {
    my ($self, $triple, $premise, $row) = @_;
    my $l = $self->{L}->val($triple, $premise, $row);
    my $regex = $self->{R}->val($triple, $premise, $row)->getString;

    my $uriWildcard = '*';
    $regex = $self->mapWildcards($regex, $uriWildcard, '\\', '.*', ['~','!','@','#',"\$",'%','^','&','*','(',')','+','`','-','=',
								    '{','}','[',']','|',':','"',';','\'','<','>','?',',','.','/']);
#    $regex =~ s/\\/\\\\/g;
#    $regex =~ s/\//\\\//g;
#    $regex =~ s/\?/\./g;
#    $regex =~ s/\*/\.\*\?/g;
    my $cmp;
    if ($l->isa('W3C::Rdf::BNode')) {
	$cmp = '['.$l->getAttribution->getUri->getUri.':'.$l->getId.']';
    } elsif ($l->isa('W3C::Rdf::String')) {
	$cmp = $l->getString;
    } elsif ($l->isa('W3C::Rdf::Uri')) {
	$cmp = $l->getUri;
    } elsif ($l->isa('W3C::Rdf::Attribution')) {
	$cmp = $l->getUri->getUri;
    } else {
	&throw(new W3C::Util::Exception(-message => "don't know how to get string value for ".ref $l));
    }
    $self->internedInt($cmp =~ m/^$regex$/);
}

sub mapWildcards {
    my ($self, $uri, $uriWildcard, $escape, $dbWildcard, $dbSpecials) = @_;

    if ($uriWildcard eq $escape) {
	$uri =~ s/\Q$escape\E/$dbWildcard/g;
    } else {
	$uri =~ s/\Q$escape\E/$escape$escape/g;
    }

    foreach my $hideMe (@$dbSpecials) {
	if ($uriWildcard ne $hideMe) {
	    $uri =~ s/\Q$hideMe\E/$escape$hideMe/g;
	}
    }

    if ($uriWildcard ne $dbWildcard && $uriWildcard ne $escape) {
	if (length($dbWildcard) == 1) {
	    $uri =~ s/\Q$dbWildcard\E/$escape$dbWildcard/g;
	}
	$uri =~ s/\Q$uriWildcard\E/$dbWildcard/g;
    }

    return $uri;
}

package W3C::Rdf::AlgaeCompileTree::Or;
sub opStr {'||';}

sub val {
    my ($self, $triple, $premise, $row) = @_;
    my $ret = undef;
    eval {
	my $r = $self->{L}->val($triple, $premise, $row);
	if ($r) {
	    if ($r->isa('W3C::Rdf::String')) {
		if ($r->getString) {
		    $ret = $r;
		}
	    } else {
		$ret = $r;
	    }
	}
    }; if ($@) {if (my $ex = &catch('W3C::Rdf::SafeEvaluationException')) {
    } else {
	&throw();
    }}

    return $ret || $self->{R}->val($triple, $premise, $row);
}

package W3C::Rdf::AlgaeCompileTree::And;
sub opStr {'&&';}
sub val {
    my $ret = $_[0]->{L}->val(@_[1..3]);
    if (!$ret || ($ret->isa('W3C::Rdf::String') && !$ret->getString)) {
	return $ret;
    } else {
	return $_[0]->{R}->val(@_[1..3]);
    }
}
sub toSQL {
    my ($self, $algae, $term, $cWheres, $outer) = @_;
    $self->{L}->toSQL($algae, $term, $cWheres, $outer);
    $self->{R}->toSQL($algae, $term, $cWheres, $outer);
}

sub W3C::Rdf::AlgaeCompileTree::Xor::opStr {'^';}
sub W3C::Rdf::AlgaeCompileTree::Xor::val {$_[0]->{L}->val(@_[1..3]) ^ $_[0]->{R}->val(@_[1..3]);}

sub W3C::Rdf::AlgaeCompileTree::Lt::opStr {'<';}
sub W3C::Rdf::AlgaeCompileTree::Lt::val {$_[0]->internedInt($_[0]->cmp($_[0]->{L}->val(@_[1..3]), $_[0]->{R}->val(@_[1..3])) < 0)}

sub W3C::Rdf::AlgaeCompileTree::Le::opStr {'<=';}
sub W3C::Rdf::AlgaeCompileTree::Le::val {$_[0]->internedInt($_[0]->cmp($_[0]->{L}->val(@_[1..3]), $_[0]->{R}->val(@_[1..3])) < 1);}

sub W3C::Rdf::AlgaeCompileTree::Gt::opStr {'>';}
sub W3C::Rdf::AlgaeCompileTree::Gt::val {$_[0]->internedInt($_[0]->cmp($_[0]->{L}->val(@_[1..3]), $_[0]->{R}->val(@_[1..3])) > 0)}

sub W3C::Rdf::AlgaeCompileTree::Ge::opStr {'>=';}
sub W3C::Rdf::AlgaeCompileTree::Ge::val {$_[0]->internedInt($_[0]->cmp($_[0]->{L}->val(@_[1..3]), $_[0]->{R}->val(@_[1..3])) > -1)}

# !!!
sub W3C::Rdf::AlgaeCompileTree::Plus::opStr {'+';}
sub W3C::Rdf::AlgaeCompileTree::Plus::val {$_[0]->internedInt($_[0]->toNum($_[0]->{L}->val(@_[1..3])) + $_[0]->toNum(($_[0]->{R}->val(@_[1..3]))))}

sub W3C::Rdf::AlgaeCompileTree::Minus::opStr {'-';}
sub W3C::Rdf::AlgaeCompileTree::Minus::val {$_[0]->internedInt($_[0]->toNum($_[0]->{L}->val(@_[1..3])) - $_[0]->toNum(($_[0]->{R}->val(@_[1..3]))))}

sub W3C::Rdf::AlgaeCompileTree::Multiply::opStr {'*';}
sub W3C::Rdf::AlgaeCompileTree::Multiply::val {$_[0]->internedInt($_[0]->toNum($_[0]->{L}->val(@_[1..3])) * $_[0]->toNum(($_[0]->{R}->val(@_[1..3]))))}

sub W3C::Rdf::AlgaeCompileTree::Exp::opStr {'^';}
sub W3C::Rdf::AlgaeCompileTree::Exp::val {$_[0]->internedInt($_[0]->toNum($_[0]->{L}->val(@_[1..3])) ** $_[0]->toNum(($_[0]->{R}->val(@_[1..3]))))}

sub W3C::Rdf::AlgaeCompileTree::Divide::opStr {'/';}
sub W3C::Rdf::AlgaeCompileTree::Divide::val {$_[0]->internedInt($_[0]->toNum($_[0]->{L}->val(@_[1..3])) / $_[0]->toNum(($_[0]->{R}->val(@_[1..3]))))}

sub W3C::Rdf::AlgaeCompileTree::Modulo::opStr {'%';}
sub W3C::Rdf::AlgaeCompileTree::Modulo::val {$_[0]->internedInt($_[0]->toNum($_[0]->{L}->val(@_[1..3])) % $_[0]->toNum(($_[0]->{R}->val(@_[1..3]))))}



sub W3C::Rdf::AlgaeCompileTree::Assign::opStr {'='}
sub W3C::Rdf::AlgaeCompileTree::Assign::getAssoc {$CalcTree::RIGHT;}
sub W3C::Rdf::AlgaeCompileTree::Assign::val {
    my ($self, $triple, $premise, $row) = @_;
    return $self->{L}->assign($triple, $premise, $row, $self->{R}->val($triple, $premise, $row));
}

package W3C::Rdf::AlgaeCompileTree::Unary;
use W3C::Util::Exception;

sub new {
    my ($proto, $M, @exprParms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@exprParms);
    $self->{M} = $M;
    return $self;
}

sub val {
    my ($self, $triple, $premise, $row) = @_;
    return "!!! $self no val defined for $self->{L} ? $self->{R}";
}

sub toString {
    my ($self, %flags) = &main::_defaultPrec(@_);
#    my $val = $self->val();
    my $opStr = $self->opStr();
    if ($flags{-precTable} && exists $flags{-precTable}{$opStr}) {
	my ($outPrec, $outAssoc) = @{$flags{-precTable}{$opStr}};
	my $mStr = $self->{M}->toString(%flags, -parentPrec => $outPrec);
	if ($outPrec < $flags{-parentPrec} || 
	    ($outPrec == $flags{-parentPrec} && $self->getAssoc == $outAssoc)) {
	    return "$opStr $mStr"; # = $val)";
	} else {
	    return "($opStr $mStr)"; # = $val)";
	}
    } else {
	my $mStr = $self->{M}->toString();
	return "($opStr $mStr)"; # = $val)";
    }
}

sub W3C::Rdf::AlgaeCompileTree::Not::opStr {'NEG';}
sub W3C::Rdf::AlgaeCompileTree::Not::val {- $_[0]->{M}->val;}
sub W3C::Rdf::AlgaeCompileTree::Not::getAssoc {$CalcTree::RIGHT;}

sub W3C::Rdf::AlgaeCompileTree::BitFlip::opStr {'~';}
sub W3C::Rdf::AlgaeCompileTree::BitFlip::val {~ $_[0]->{M}->val;} # !!!
sub W3C::Rdf::AlgaeCompileTree::BitFlip::getAssoc {$CalcTree::RIGHT;}

sub W3C::Rdf::AlgaeCompileTree::Lang::opStr {'!!!';}
sub W3C::Rdf::AlgaeCompileTree::Lang::val {$_[0]->{M}->getLang();}

sub W3C::Rdf::AlgaeCompileTree::Label::opStr {'!!!';}
sub W3C::Rdf::AlgaeCompileTree::Label::val {
    my ($self, $triple, $premise, $row) = @_;
    my $val = $self->{M}->val($triple, $premise, $row);
    if ($val->isa('W3C::Rdf::String')) {
	return $val;
    } else {
	&throw(new W3C::Rdf::FailureOnTypeException(-expect => 'W3C::Rdf::String', 
						    -got => ref $val));
    }
}

1;

__END__

=head1 NAME

W3C::Rdf::AlgaeCompileTree - RDF query objects

=head1 SYNOPSIS

  # in a parser...
  use W3C::Rdf::AlgaeCompileTree;
  my $t1 = new W3C::Rdf::AlgaeCompileTree::Decl([$p1, $s1, $o1], 
				$constraints1, $self);
  my $t2 = new W3C::Rdf::AlgaeCompileTree::Decl([$p2, $s2, $o2], 
				$constraints2, $self);
  my $conj = new W3C::Rdf::AlgaeCompileTree::Conjunction($t1, $t2, $self)

=head1 DESCRIPTION

Parsers generate AlgaeCompileTrees and hand them to the Algae2 object. Algae2
calls the compile tree to perforn queries/assertions/rules.

This module is part of the W3C::Rdf CPAN module.

=head1 METHODS

    $atomDictionary, $namespaceHandler, 
	$sources, $rdfApp, $sourceAttrib, $flags

=head2 immediateEvaluate($resultSet)

Perform compile-time evaluations.

no return value

=head2 delayedEvaluate($resultSet)

Perform post-compile-time (pass 2) evaluations. This is for query languages
that require a second pass to do things like resolve namespace prefixes that
are declared after a dependent qname is first used.

no return value

=head2 val($triple, $premise, $row)

Perform the query evaluation. I don't remember what the hell the $premise is,
but it sure looks important.

returns an interned RDF Atom

=head1 AUTHOR

Eric Prud'hommeaux <eric@w3.org>

=head1 SEE ALSO

W3C::Rdf::Algae2(1) W3C::Rdf::AlgaeParser(1) W3C::Rdf::Atom(1) perl(1).

=cut
