#Copyright Massachusetts Institute of technology, 1998.
#Written by Eric Prud'hommeaux for the World Wide Web Consortium

# $Id: RdfApp.pm,v 1.102 2004/06/12 12:50:25 eric Exp $

use strict;
require Exporter;
require AutoLoader;

$W3C::Rdf::RdfApp::REVISION = '$Id: RdfApp.pm,v 1.102 2004/06/12 12:50:25 eric Exp $ ';


package W3C::Rdf::RdfApp;
use vars qw($VERSION $DSLI @ISA @EXPORT @EXPORT_OK);
@ISA = qw(W3C::Util::NamedParmObject Exporter AutoLoader);

use W3C::Rdf::Parser;
use W3C::XML::XmlParser;
use W3C::XML::InputSource;
use W3C::XML::HandlerBase;
use W3C::XML::ErrorHandlerImpl;
use W3C::Rdf::Atoms qw($RDF_SCHEMA_URI $ATTRIB_GroundFact);
use W3C::Rdf::RdfDB;
use W3C::Util::Exception;
@EXPORT = qw();
@EXPORT_OK = qw();
$VERSION = 0.94;
$DSLI = 'adpO';

use vars qw($FLAG $LOOK_FOR $STORE_IN $DESCRIPTION $DEFAULT $OPTIONS $TYPE_SPECIFIC);
($FLAG, $LOOK_FOR, $STORE_IN, $DESCRIPTION, $DEFAULT, $OPTIONS, $TYPE_SPECIFIC) = (0..6);


#####
# Externally available constants
*W3C::Rdf::RdfApp::ONE = \ 'one';

sub new {
    my ($proto, $ARGV, @parms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@parms);
    bless ($self, $class);
    $self->{-atomDictionary} = new W3C::Rdf::Atoms() if (!$self->{-atomDictionary});
    $self->setInputAttribution($0);
    $self->{ARGS}{'-inputClass'} = {};
    $self->{ARGS}{'-inputParms'} = {};
    $self->{ARGS}{'-serializationClass'} = {};
    $self->{ARGS}{'-serializationParms'} = {};
    $self->{CMND_SET} = [
	[0, '-inputLang', \ $self->{ARGS}{'-inputLanguage'}, 'non-XML input language', undef, [qw(n3 <class name>)]], 
	[0, '-inputParm', sub {my ($parm) = @_;
			       if ($parm =~ m/([^= ]+) = (.*)/x) {
				   $self->{ARGS}{-inputParms}{$1} = $2;
			       } else {
				   $self->{ARGS}{-inputParms}{$parm} = 1;
			       }}, 'set parser variable', undef], 
	[0, '-rdfns', \ $self->{ARGS}{'-rdfUri'}, 'rdf model and syntax URI', 
	 $RDF_SCHEMA_URI], 
	[0, '-rule', \ @{$self->{ARGS}{'-rules'}}, 'rule assertion files', []], 
	[1, '-i', \ $self->{ARGS}{'-interactiveQueryBuilder'}, 'Use Interactive Query Builder', undef, undef, [1]], 
	[0, '-ns', \ @{$self->{ARGS}{'-namespace'}}, 'namespace declarations', []], 
	[0, '>', sub {open (OUT, '>'.shift) || die $!; $self->{OUT_FILE} = *OUT;}, 'output to file', undef], 
	[0, '-d', \ $self->{ARGS}{'-rdfDB'}, 'rdf database', '<http://www.w3.org/1999/02/26-modules/algae#ephemeral> ()'], 
	[0, '-p', \ $self->{ARGS}{'-parser'}, 'XML parser', 'W3C::XML::PerlXmlParser', ['W3C::XML::ExpatXmlParser']], 
	[0, '-set', sub {my ($parm) = @_;
			 if ((my ($attr, $value) = split ('\s*=\s*', $parm)) == 2) {
			     $self->{ARGS}{-props}{$attr} = $value;
			 } else {
			     push (@{$self->{ERRORS}}, "malformed set command: $parm");
			 }}, 'set variable', undef], 
	[0, '-sClass', \ $self->{ARGS}{'-serializationClass'}, 'how to show results', undef, [qw(RDFXML dot arrows n3 ntriples statements <class name>)]], 
	[0, '-sParm', sub {my ($parm) = @_;
			   if ($parm =~ m/([^= ]+) = (.*)/x) {
			       $self->{ARGS}{-serializationParms}{$1} = $2;
			   } else {
			       $self->{ARGS}{-serializationParms}{$parm} = 1;
			   }}, 'set variable', undef], 
	[1, '-so', \ $self->{ARGS}{'-sourceOnly'}, 'source only', undef, ['source only'], [1, 'source only']], 
	[0, '-o', \ $self->{ARGS}{'-outputFile'}, 'output file', undef], 
	[1, '-RDFObject', \ $self->{ARGS}{'-startState'}, 'RDF grammer with no RDF decl', 
	 undef, ['RDFObject'], ['RDFObject', 'RDFObject']], 
	[1, '-ByPredicate', \ $self->{ARGS}{'-startState'}, 'ByPredicate grammer', 
	 undef, ['ByPredicate'], ['ByPredicate', 'ByPredicate']], 
	[1, '-BySubjectType', \ $self->{ARGS}{'-startState'}, 'BySubjectType grammer', 
	 undef, ['BySubjectType'], ['BySubjectType', 'BySubjectType']], 
	[1, '-BySubject', \ $self->{ARGS}{'-startState'}, 'BySubject grammer', 
	 undef, ['BySubject'], ['BySubject', 'BySubject']], 
	[0, '-a', \ @{$self->{ARGS}{'-algae'}}, 'algae queries', []], 
	[0, '-r', \ @{$self->{ARGS}{'-rdql'}}, 'rdql queries', []], 
	[0, '-s', \ @{$self->{ARGS}{'-serql'}}, 'serql queries', []], 
    ];

    return $self;
}

sub setInputAttribution {
    my ($self, $urlStr) = @_;
    my ($hostname, $cwd) = ('localhost', '.');
    eval {
	require Sys::Hostname;
	$hostname = &Sys::Hostname::hostname;
    };
    eval {
	use Cwd;
	$cwd = &getcwd;
	chomp $cwd;
	if ($^O eq 'MSWin32' && $cwd =~ m/^([A-Z])\:(.*)$/) {
	    $cwd = '/'.$1.'%3A'.$2;
	}
    };
    my $base = new URI("file://$hostname$cwd/");
    my $url = $self->{-atomDictionary}->getUri(new_abs URI($urlStr, $base));
    $self->{INPUT_ATTRIBUTION} = $self->{-atomDictionary}->
	getGroundFactAttribution($url, undef, undef, undef);
}

sub addCommands (\@) {
    my ($self, $command) = @_;
    push (@{$self->{CMND_SET}}, $command);
}

sub execute (\@) {
    my ($self, $argv) = @_;
    $self->preArgumentDefaults($argv);
    $self->parseArgumentArray($argv);
    $self->postArgumentDefaults($argv);
    if ($self->{ARGS}{'-help'}) {
	$self->showHelp(*STDOUT);
    } else {
	$self->parse;
	$self->render;
    }
}

sub preArgumentDefaults {
    my ($self) = @_;
    $self->{ARGS}{-props} = {};
    $self->assignDefaults;
    return 0;
}

sub parseArgumentArray {
    my ($self, $argv) = @_;
  NEXTARG:
    for (my $i = 0; $i < @$argv; $i++) {
	my $arg = $argv->[$i];
	$self->{-location} = "cmdline-arg:$i/" if (!$self->{-location} || $self->{-location} =~ m/^cmdline-arg:/);
	if ($arg =~ m/\A @(.*) \Z/x) {
	    my $scriptFile;
	    if ($1) {
		$scriptFile = $1;
	    } elsif ($i < @$argv - 1) {
		$scriptFile = $argv->[++$i];
	    } else {
		push (@{$self->{ERRORS}}, 'missing parm for \@ directive');
	    }
	    $self->{-location} = "file:$scriptFile";
	    if (!open (SCRIPT_FILE, '<'.$scriptFile)) {
		&throw(new W3C::Util::FileNotFoundException(-filename => $scriptFile));
	    }
	    local $/ = undef;
	    my $script = <SCRIPT_FILE>;
	    close(SCRIPT_FILE);
#	    my $scriptUri = $scriptFile =~ m/^\// ? "file://$scriptFile" : "file:$scriptFile";
	    $self->setInputAttribution($scriptFile);
	    my ($line, @parms);
	    while (pos $script < length $script && $script =~ m/\G ([^\s\\\'\"\#]*) ([\s\\\'\"\#])? /gcxs) {
		my ($word, $delim) = ($1, $2);
		$line .= $word;
		if ($delim =~ m/\s/) {
		    push (@parms, $line) if ($line && $line ne '$*');
		    $line = undef;
		    if ($script =~ m/\G \s* /gcxs) {}
		} elsif ($delim eq '\\') {
		    if ($script =~ m/\G (.) /gcxs) {$line .= $1 if ($1 ne "\r" && $1 ne "\n");}
		} elsif ($delim eq '\'') {
		    while ($script =~ m/\G ([^\\\']*) ([\\\']) /gcxs) {
			$line .= $1;
			if ($2 eq '\\') {
			    if ($script =~ m/\G (.) /gcxs) {$line .= $1 if ($1 ne "\r" && $1 ne "\n");}
			} elsif ($2 eq '\'') {
			    last;
			}
		    }
		} elsif ($delim eq '"') {
		    while ($script =~ m/\G ([^\\\"]*) ([\\\"]) /gcxs) {
			$line .= $1;
			if ($2 eq '\\') {
			    if ($script =~ m/\G (.) /gcxs) {$line .= $1 if ($1 ne "\r" && $1 ne "\n");}
			} elsif ($2 eq '"') {
			    last;
			}
		    }
		} elsif ($delim eq '#') { # eat rest of line after comment
		    $script =~ m/\G ([^\r\n]*) $ /gcxm;
		} else {
		    push (@parms, $line) if (defined $line && $line ne '$*');
		    $line = undef;
		    # \z so at end of input - hmm - can't count on that. bummer
		}
	    }
	    push (@parms, $line) if ($line);
	    my $zero = shift (@parms);
	    splice (@$argv, $i, 1, @parms);
	    $i--;
	} elsif ($self->matchInput($arg, $argv, \$i)) {
	} elsif ($arg =~ m/\A ( -{0,2}help | -{0,1}\? ) \Z/x) { # help -help --help ? -?
	    $self->{ARGS}{'-help'} = 1;
	} elsif ($arg =~ m/\A ( -{1,2}verbose | -{1,2}v ) \Z/x) { # -verbose --verbose -v --v
	    $self->{ARGS}{'-verbose'} = 1;
	} else {
	    push (@{$self->{ARGS}{-files}}, $arg);
	}
    }
}

sub assignDefaults {
    my ($self) = @_;
    for (my $cmdNo = 0; $cmdNo < @{$self->{CMND_SET}}; $cmdNo++) {
	my $cmnd = $self->{CMND_SET}[$cmdNo];
	my $dflt = $cmnd->[$DEFAULT];
	if (ref $cmnd->[$STORE_IN] eq 'CODE') {
	} elsif (ref $cmnd->[$STORE_IN] eq 'ARRAY') {
	    push (@{$cmnd->[$STORE_IN]}, @$dflt);
	} elsif (ref $cmnd->[$STORE_IN] eq 'HASH') {
	    %{$cmnd->[$STORE_IN]} = %$dflt;
	} else {
	    $ {$cmnd->[$STORE_IN]} = $dflt;
	}
    }
    return 0;
}

sub matchInput {
    my ($self, $arg, $argv, $iPtr) = @_;
    for (my $cmdNo = 0; $cmdNo < @{$self->{CMND_SET}}; $cmdNo++) {
	my $cmnd = $self->{CMND_SET}[$cmdNo];
	my $typeSpecific = $cmnd->[$TYPE_SPECIFIC];
	if ($arg =~ m/\A $cmnd->[$LOOK_FOR](.*) \Z/xs) {
	    if ($cmnd->[$FLAG]) {
		$ {$cmnd->[$STORE_IN]} = $typeSpecific->[0];
	    } else {
		my $value = $1;
		if ($value) {
		} elsif ($$iPtr < @$argv - 1) {
		    $value = $argv->[++$$iPtr];
		} else {
		    push(@{$self->{ERRORS}}, "unknown arg: $arg: \"$value\"");
		}
		if ($value) {
		    if (ref $cmnd->[$STORE_IN] eq 'ARRAY') {
			push (@{$cmnd->[$STORE_IN]}, $value);
		    } elsif (ref $cmnd->[$STORE_IN] eq 'HASH') {
			$cmnd->[$STORE_IN]{$value} = $value;
		    } elsif (ref $cmnd->[$STORE_IN] eq 'CODE') {
			my $call = $cmnd->[$STORE_IN];
			&$call($value);
		    } else {
			$ {$cmnd->[$STORE_IN]} = $value;
		    }
		}
	    }
	    return $cmdNo + 1;
	}
    }
    return 0;
}

sub postArgumentDefaults {
    my ($self) = @_;
    return 0;
}

sub showHelp {
    my ($self, $sink) = @_;
    print $sink "Usage: $0 options:\n";
    for (my $cmdNo = 0; $cmdNo < @{$self->{CMND_SET}}; $cmdNo++) {
	my $cmnd = $self->{CMND_SET}[$cmdNo];
	my $isAFlag = $cmnd->[$FLAG] ? 'f' : ' ';
	my $lookFor = $cmnd->[$LOOK_FOR];
	my $desc = $cmnd->[$DESCRIPTION];
	my $dflt = ref $cmnd->[$DEFAULT] eq 'ARRAY' ? @{$cmnd->[$DEFAULT]} : $cmnd->[$DEFAULT];
	my $value = ref $cmnd->[$STORE_IN] eq 'ARRAY' ? join ('|', @{$cmnd->[$STORE_IN]}) : 
	    ref $cmnd->[$STORE_IN] eq 'CODE' ? '' : 
		$ {$cmnd->[$STORE_IN]};
	if ($cmnd->[$FLAG] && $value != $dflt) {
	    $value = $cmnd->[$TYPE_SPECIFIC][1];
	}
	my @opts;
	@opts = @{$cmnd->[$OPTIONS]} if ($cmnd->[$OPTIONS]);
	unshift (@opts, "[$dflt]") if ($dflt);
	my $optionString;
	$optionString = ' ('.join ('|', @opts).')' if (@opts);
	$value = " = $value" if ($value);
	print $sink "$lookFor - $desc$optionString$value\n";
    }
    my $sources = $self->{ARGS}{-files} ? join (', ', @{$self->{ARGS}{-files}}) : '';
    if ($sources) {
	print $sink "reading from $sources\n";
    } else {
	print "plus a list of files or URLs from which to read RDF\n";
    }
}

sub altParser {
    my ($self);
    my $parserClass = $self->{ARGS}{-inputLanguage};
    if (defined $testRdfParser::ParserClassAbbreviations{$parserClass}) {
	$parserClass = $testRdfParser::ParserClassAbbreviations{$parserClass};
    }
    if ($parserClass) {
	# Put the flags together.
	my $flags = {-indent => 3, 
#		     -publicId => $source, 
#		     -systemId => $source, 
		     -noRdfTag => 0, 
		     -RdfDB => $self->{RDF_DB}, 
		     -allowAnonymousRefs => 1, 
		     -ignoreReifications => 1, 
		     -ignoreAnonymousReifications => 1, 
		     -neededNamespaces =>[qw(r)],
		     -atomDictionary => $self->{-atomDictionary}, 
#		     -resource => $outAttribution, 
#		     -createNamespaces => 1, 
		     -importMap => $self->{NAMESPACE_HANDLER}->getAggregateNamespaceHandler()};

	# Incorporate parser parms.
	$flags = {%$flags, %{$self->{ARGS}{-parserParms}}};

	# Build the parser class.
	my $parser;
	eval "require $parserClass; \$parser = new $parserClass(\$flags);";
	if ($@) {&throw()}
    }

}

sub prepareParser () {
    my ($self) = @_;
    if ($self->{ARGS}{-parser} eq 'W3C::XML::ExpatXmlParser') {
	require XML::Parser;
	$self->{XML_PARSER} = new W3C::XML::ExpatXmlParser;
    } else {
	$self->{XML_PARSER} = new W3C::XML::PerlXmlParser;
    }
#    $self->{XML_PARSER} = new W3C::XML::PerlXmlParser;
#    $self->{XML_PARSER} = new W3C::XML::ExpatXmlParser;
    my $errorHandler = new W3C::Rdf::RdfDbErrorHandler(-documentLocator => $self->{XML_PARSER});

    # pick an RDF database
    if (my $dbSpec = $self->{ARGS}{-rdfDB}) {
	# do we need this interface anyays?
	# if so, create an AlgaeParser and ask it to parse "attach $self->{ARGS}{-rdfDB}"
	$self->{RDF_DB} = new W3C::Rdf::RdfDB(-atomDictionary => $self->{-atomDictionary}, -errorHandler => $errorHandler, -lazyReification => 1);
    } else {
	require W3C::Rdf::RdfDB;
	$self->{RDF_DB} = new W3C::Rdf::RdfDB(-atomDictionary => $self->{-atomDictionary}, -errorHandler => $errorHandler, -lazyReification => 1);
    }

    $self->{RDF_PARSER} = new W3C::Rdf::Parser(-atomDictionary => $self->{-atomDictionary}, 
					       -nextTrustControlBlock=>1, -trustedIds=>{}, -untrustedIds=>{}, 
					       -rdfDB=>$self->{RDF_DB}, -namespaceHandler=>undef, 
					       -errorHandler=>$errorHandler, -appendAttribution=>0);

    # use a W3C::XML::NamespaceAndStringHandler stream to map from prefix:tag - pass to rdfParser
    # so it translate back for pretty-printing
    $self->{NAMESPACE_HANDLER} = new W3C::XML::NamespaceAndStringHandler(-documentHandler => $self->{RDF_PARSER}, 
									 -errorHandler => $errorHandler, 
									 -passUnknownNamespaces => 1, -useSAX2 => 1, 
									 -failAttrNStoEltNs => $self->{ARGS}{-props}{-failAttrNStoEltNs});
    $self->{RDF_PARSER}->setNamespaceHandler($self->{NAMESPACE_HANDLER});

    # point the XML Parser at the top of the DocumentHandler stack.
    $self->{XML_PARSER}->setDocumentHandler($self->{NAMESPACE_HANDLER});
    $self->{XML_PARSER}->setErrorHandler($errorHandler);
}

sub parse () {
    my ($self, $parsables) = @_;

    if (!$self->{RDF_PARSER}) {
	$self->prepareParser();
    }

    if (!$parsables) {
	$parsables = $self->{ARGS}{-files};
    }

    # assign a URI to the XML in InputSource and parse it
    foreach my $source (@$parsables) {
	my $inputSource;
	eval {
	    $inputSource = &getInputSource($source);
	    $self->parseOne($inputSource);
	}; if ($@) {if (my $ex = &catch('W3C::Util::FileNotFoundException')) {
	    warn "Can't open input file \"".$ex->getFilename."\" - skipping\n";
	} elsif ($ex = &catch('W3C::XML::SAXParseException')) {
	    # SAXParseExceptions are not thrown by getInputSource.
	    warn $ex->toString.' - abandoning "'.$inputSource->getPublicId."\".\n";
	} elsif ($ex = &catch('W3C::Util::Exception')) {
	    print $ex->toString;
	    exit;
	} else {die $@;}}
    }
}

sub getInputSource { # static
    my ($source, $parms) = @_;
    my $ret = undef;
    my $url = new URI($source);
    if ($url->scheme) {
	eval {
	    if ($url->scheme() eq 'file') {
		$ret = new W3C::XML::FileInputSource($url->path(), $parms);
	    } else {
		$ret = new W3C::XML::URIInputSource($url->canonical, $parms);
	    }
	}; if ($@) {if (my $ex = &catch('W3C::Util::LibraryNotFoundException')) {
	    if ($source =~ m/^file\:(.+)/) {
		$ret = new W3C::XML::FileInputSource($1);
	    } else {
		&throw($ex);
	    }
	} else {&throw()}}
    } else {
	$ret = new W3C::XML::FileInputSource($source);
    }
    return $ret;
}

sub getRdfDB {
    my ($self) = @_;
    return $self->{RDF_DB};
}

sub setRdfDB {
    my ($self, $rdfDB) = @_;
    $self->{RDF_DB} = $rdfDB;
    $self->{RDF_PARSER}->setRdfDB($rdfDB);
}

sub parseOne {
    my ($self, $inputSource, $type) = @_;
    my $pubId = $inputSource->getPublicId;
    if ($type eq 'n3') {
	require RDF::Notation3;
	require RDF::Notation3::Triples;
	my $rdf = RDF::Notation3::Triples->new();
	$rdf->parse_string($inputSource->getByteStream);
	my $triples = $rdf->get_triples;
	my $baseUri = $self->{-atomDictionary}->getUri($pubId);
	my ($auth, $parent, $source, $trusted, $clear) = (undef, undef, $baseUri, 1, 1);
	my $attrib = $self->{-atomDictionary}->getAttribution($ATTRIB_GroundFact, $baseUri, $auth, 
							      $parent, $source, $trusted, $clear);

	my $anonMap = {};
	my %spoc2psorbc = (0=>1, 1=>0, 2=>2, 3=>5);
	foreach my $row (@$triples) {
	    next if ($row->[1] eq '<http://www.w3.org/2000/10/swap/log.n3#forSome>');
	    my @rdfIds = (undef, undef, undef, undef, undef, undef);
	    for (my $i = 0; $i < @$row; $i++) {
		my $object = $row->[$i];
		my $interned;
		if ($object =~ m/^\<\>$/) {
		    $interned = $i == 3 ? $attrib : $attrib->getUri;
		} elsif ($object =~ m/^[\<]\#g_(\d+)[\>]$/) {
		    my $idNo = $1;
		    if (!($interned = $anonMap->{$idNo})) {
			$interned = $anonMap->{$idNo} = $self->{-atomDictionary}->createBNode($attrib);
		    }
		} elsif ($object =~ m/^[\<]([^\>]+)[\>]$/) {
		    $interned = $self->{-atomDictionary}->getUri($1);
		} elsif ($object =~ m/^[\"]([^\"]+)[\"]$/) {
		    $interned = $self->{-atomDictionary}->getString($1, undef, 'PLAIN');
		} else {
		    &throw(new W3C::Util::ProgramFlowException);
		}
		$rdfIds[$spoc2psorbc{$i}] = $interned;
	    }
#	    eval {
		$self->{RDF_DB}->addStatement(@rdfIds);
#	    }; if ($@) {
#		print '!',join(' ', @$row),"\n";
#	    } else {
#		print ' ',join(' ', @$row),"\n";
#	    }
	}
    } else {
	$self->{NAMESPACE_HANDLER}->addResourceMap($pubId);
	$self->{RDF_PARSER}->setState($self->{ARGS}{'-startState'}) if (defined $self->{ARGS}{'-startState'});
	$self->{XML_PARSER}->parse($inputSource);
	$self->{ATTRIBUTIONS}{$pubId} = $self->{RDF_PARSER}->getRootAttribution;
	# print join (' | ', ($self->{NAMESPACE_HANDLER}->getUnknownNamespaces))."\n";
	if (defined $self->{NAMESPACE_HANDLER}->getUnknownNamespaceDeclaration('rdf')) {
	    my $rdfNs = $RDF_SCHEMA_URI;
	    my $unknown = $self->{NAMESPACE_HANDLER}->getUnknownNamespaceDeclaration('rdf');
	    warn "rdf: namespace undeclared: assuming $rdfNs
-- fix this by adding 'xmlns:rdf=\"$rdfNs\"' to rdf:$unknown\n";
	    $self->{NAMESPACE_HANDLER}->addNamespace('rdf', $rdfNs, $pubId);
	}
	$self->{RDF_PARSER}->append;
    }
    return $pubId;
}

sub render () {
}

1;

__END__

=head1 NAME

W3C::Rdf::RdfApp - Common handlers and CLI for RDF apps

=head1 SYNOPSIS

    use W3C::Rdf::RdfApp;
    package testRdfParser;
    @testRdfParser::ISA = qw(W3C::Rdf::RdfApp);

    my $tester = new testRdfParser;
    $tester->execute(\@ARGV);

    sub render {
	my ($self) = @_;
	$query = "(ask '((http://www.w3.org/schema/certHTMLv1/access ?x \"acc1\") 
(?y ?x http://user1))
:collect ?x)" if (!$query);
	my $queryHandler = $self->{RDF_DB}->getAlgaeInterface;
	my ($nodes2, $messages) = $queryHandler->algae($query, $self->{RDF_PARSER}->{REVISION});
    }

=head1 DESCRIPTION

Common handlers and CLI for RDF apps.

This module is part of the W3C::RDF CPAN module suite.

=head1 ARGUMENTS

=item B<@<file>> - read arguments from <file>

=item B<...>

=item B<>

=item B<>

=item B<>

=head1 AUTHOR

Eric Prud\'hommeaux <eric@w3.org>

=head1 SEE ALSO

W3C::Rdf::Parser(3) W3C::Rdf::RdfDB(3) W3C::XML::XmlParser(3) perl(1).

=cut
