#Copyright Massachusetts Institute of technology, 1998.
#Written by Eric Prud'hommeaux for the World Wide Web Consortium

# $Id: ResultSet.pm,v 1.70 2004/06/16 05:12:43 eric Exp $

use strict;
require Exporter;
#require AutoLoader;

$W3C::Rdf::ResultSet::REVISION = '$Id: ResultSet.pm,v 1.70 2004/06/16 05:12:43 eric Exp $ ';

package W3C::Rdf::Result;
use W3C::Util::Exception;

sub new {
    my ($proto, $resultSet, $row, $iterator) = @_;
    my $class = ref($proto) || $proto;
    my $self = {RESULT_SET => $resultSet, ROW => $row, ITERATOR => $iterator};
    bless ($self, $class);
    return $self;
}
sub get {
    my ($self, $column) = @_;
    return $self->{RESULT_SET}{RESULTS}[$self->{ROW}][$column];
}
sub set {
    my ($self, $column, $value) = @_;
    $self->{RESULT_SET}{RESULTS}[$self->{ROW}][$column] = $value;
}
sub duplicate {
    my ($self) = @_;
    # splice in at current index and increment indexes
    if (@{$self->{RESULT_SET}{RESULTS}} <= $self->{ROW}) {
	&throw(new W3C::Util::OutOfBoundsException(-name => 'Result', -index => $self->{ROW}));
    }
    my $newRow = [@{$self->{RESULT_SET}{RESULTS}[$self->{ROW}]}];
    my $newProofs = $self->{RESULT_SET}->_makeDB;
    $newProofs->copyTriples($self->{RESULT_SET}{STATEMENTS}[$self->{ROW}]);
    my $newOb = new W3C::Rdf::Result($self->{RESULT_SET}, $self->{ROW}, $self->{ITERATOR});
    splice (@{$self->{RESULT_SET}{RESULTS}}, $self->{ROW}, 0, $newRow);
    splice (@{$self->{RESULT_SET}{STATEMENTS}}, $self->{ROW}, 0, $newProofs);
    ++$self->{ROW};
    ++$self->{ITERATOR}{CUR_ROW};
    return $newOb;
}
sub eliminate {
    my ($self) = @_;
    # splice out current and decrement indexes
    splice (@{$self->{RESULT_SET}{RESULTS}}, $self->{ROW}, 1);
    splice (@{$self->{RESULT_SET}{STATEMENTS}}, $self->{ROW}, 1);
    # --$self->{ROW};
    --$self->{ITERATOR}{CUR_ROW};
}
# unduplicate - remove duplicated rows that weren't needed.
sub unduplicate {
    my ($self, $doomed) = @_;
    $doomed->eliminate();
    --$self->{ROW};
}
sub addProofs {
    my ($self, $proofs) = @_;
    foreach my $proof (@$proofs) {
	$self->{RESULT_SET}{STATEMENTS}[$self->{ROW}]->addTriple($proof);
    }
}
sub getProofs {
    my ($self) = @_;
    return $self->{RESULT_SET}{STATEMENTS}[$self->{ROW}]->getTriples;
}
sub getBindings {
    my ($self) = @_;
    my %ret;
    for (my $col = 0; $col < @{$self->{RESULT_SET}{INDEX_VAR}}; $col++) {
	$ret{$self->{RESULT_SET}{INDEX_VAR}[$col]} = $self->{RESULT_SET}{RESULTS}[$self->{ROW}][$col];
    }
    return %ret;
}
sub makeResultSet {
    my ($self) = @_;
    my $ret = {%{$self->{RESULT_SET}}};
    $ret->{RESULTS} = [[@{$self->{RESULT_SET}{RESULTS}[$self->{ROW}]}]];
    my $newProofs = $self->{RESULT_SET}->_makeDB;
    $newProofs->copyTriples($self->{RESULT_SET}{STATEMENTS}[$self->{ROW}]);
    $ret->{STATEMENTS} = [$newProofs];
    bless ($ret, 'W3C::Rdf::ResultSet');
    return $ret;
}
sub assumeNewBindings {
    my ($self, $copyFrom) = @_;
    my $dLen = @{$self->{RESULT_SET}{RESULTS}[$self->{ROW}]};
    my $newDLen = @{$copyFrom->{RESULT_SET}{RESULTS}[$copyFrom->{ROW}]};
    my @sTriples = $self->{RESULT_SET}{STATEMENTS}[$self->{ROW}]->getTriples;
    my @cfTriples = $copyFrom->{RESULT_SET}{STATEMENTS}[$copyFrom->{ROW}]->getTriples;
    for (my $i = $dLen; $i < $newDLen; $i++) {
	$self->set($i, $copyFrom->get($i));
    }
    foreach my $cfTriple (@cfTriples) {
	if (! grep {$_ == $cfTriple} @sTriples) {
	    $self->addProofs([$cfTriple]);
	}
    }
    $self->{RESULT_SET}->debugCheck;
}
sub assumeNewProofs {
    my ($self, $copyFrom) = @_;
    my @sTriples = $self->{RESULT_SET}{STATEMENTS}[$self->{ROW}]->getTriples;
    my @cfTriples = $copyFrom->{RESULT_SET}{STATEMENTS}[$copyFrom->{ROW}]->getTriples;
    foreach my $cfTriple (@cfTriples) {
	if (! grep {$_ == $cfTriple} @sTriples) {
	    $self->addProofs([$cfTriple]);
	}
    }
    $self->{RESULT_SET}->debugCheck;
}
sub getCousinWithBindings {
    my ($self, $copyFrom) = @_;
    my $width = @{$copyFrom->{RESULT_SET}{RESULTS}[$copyFrom->{ROW}]};
    my $ret = undef;
    for (my $rowNo = 0; $rowNo < @{$self->{RESULT_SET}{RESULTS}}; $rowNo++) {
	my $matches = 1;
	if (@{$self->{RESULT_SET}{RESULTS}[$rowNo]} != $width) {
	    last;
	}
	for (my $i = 0; $i < $width; $i++) {
	    if ($self->{RESULT_SET}{RESULTS}[$rowNo][$i] != $copyFrom->get($i)) {
		$matches = 0;
		last;
	    }
	}
	if ($matches) {
	    $ret = new W3C::Rdf::Result($self->{RESULT_SET}, $rowNo, $self->{ITERATOR});
	    last;
	}
    }
    return $ret;
}
sub toString {
    my ($self, %flags) = @_;
    return $self->{RESULT_SET}->toString(%flags, -markRow => $self->{ROW}, -rowMarker => "row:$self->{ROW}");
}

package W3C::Rdf::ResultSetIterator;
sub new {
    my ($proto, $resultSet) = @_;
    my $class = ref($proto) || $proto;
    my $self = {RESULT_SET => $resultSet, CUR_ROW => -1};
    bless ($self, $class);
    return $self;
}
sub hasMoreElements {
    my ($self) = @_;
    return $self->{CUR_ROW} < @{$self->{RESULT_SET}{RESULTS}}-1; # || $self->{RESULT_SET}{UNUSED};
}
sub hasMoreResults {
    my ($self) = @_;
    return $self->{CUR_ROW} < @{$self->{RESULT_SET}{RESULTS}}-1;
}
sub nextElement {
    my ($self) = @_;
    return new W3C::Rdf::Result($self->{RESULT_SET}, ++$self->{CUR_ROW}, $self);
}

package W3C::Rdf::ResultSet;
@W3C::Rdf::ResultSet::ISA = qw(W3C::Util::NamedParmObject);
use W3C::Util::Exception;
use W3C::Rdf::Atoms qw($Value_NULL);
#use W3C::Util::SExpr;

sub new {
    my ($proto, @parms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@parms);
#    $self->{-sources} ||= [];
    if (!$self->{-atomDictionary}) {
	&throw(new W3C::Util::Exception(-message => "$class must have an -atomDictionary"));
    }
    if (!UNIVERSAL::can($self->{-atomDictionary}, 'getUriLink')) {
	&throw(new W3C::Util::Exception(-message => "$class must have an -atomDictionary (not $self->{-atomDictionary})"));
    }
    $self->clear;
    return $self;
}

sub clear {
    my ($self) = @_;
    $self->{RESULTS} = [[]];	# Matrix of node results.
    $self->{STATEMENTS}= [$self->_makeDB];	# Supporting statements for each row of results.
    $self->{VAR_INDEX} = {};	# RESULTS column for a give variable (or const).
    $self->{INDEX_VAR} = [];	# Variable (or const) assigned to each column.
#    $self->{REPORT_OFFSET} = 0;	# How many columns to skip for previous collects

    # Report stuff
    $self->{UNUSED} = 1;
    $self->{SELECTS} = [];
    $self->{ITERATORS} = [];
}

sub _makeDB {new W3C::Rdf::RdfDB(-atomDictionary => $_[0]->{-atomDictionary}, -lazyReification => 1);}

# Make sure the result set has at least one row. This is useful when a previous
# query has eliminated the last row from the result set. It then seems to make
# sense to let the unification begin again.
sub atLeastOneRow {
    my ($self) = @_;
    if (!@{$self->{RESULTS}}) {
	$self->{RESULTS} = [[]];
	$self->{STATEMENTS}= [$self->_makeDB];
    }
}
sub isFresh {
    my ($self) = @_;
    return @{$self->{RESULTS}} == 1 && @{$self->{RESULTS}[0]} == 0;
}

# RdfDB interface to ResultSet -- used by RdfDB::processQueries
sub getResults {($_[0]->{RESULTS}, $_[0]->{STATEMENTS})}
sub setResultsNUKE {$_[0]->{RESULTS} = $_[1]; $_[0]->{STATEMENTS} = $_[2];}

# clearResults - Called when someone else knows that the constraints
#                aren't going to be met.
sub clearResults {
    my ($self) = @_;
    ($self->{RESULTS}, $self->{STATEMENTS}, $self->{UNUSED}) = ([], [], 1);
}

# $baseUri	URI from which relative URIs are referenced
# $flags	special control flags for this opperation

sub buildQuerySets {
    my ($self, $pattern, $messages, $db, $algae, $namespaceHandler, $baseUri, $flags) = @_;
    my $ret = new W3C::Rdf::Algae::SetQueryPiece(-999, $pattern->[0], 0, 0);
    $self->_parseExprs($ret, $pattern, $messages, $db, $algae, $namespaceHandler, $baseUri, $flags);
    return $ret;
}

sub bindSymbol {
    my ($self, $symbol) = @_;
    my $index = @{$self->{INDEX_VAR}};
    push (@{$self->{INDEX_VAR}}, $symbol);
    $self->{VAR_INDEX}{$symbol} = $index;
    return $index;
}

sub getSymbols {
    my ($self) = @_;
    return @{$self->{INDEX_VAR}};
}

sub getIndex {
    my ($self, $symbol) = @_;
    return exists $self->{VAR_INDEX}{$symbol} ? $self->{VAR_INDEX}{$symbol} : undef;
}

sub addSelect {
    my ($self, $select) = @_;
#    $self->{REPORT_OFFSET}++;
    push (@{$self->{SELECTS}}, $select);
}
sub gatherCollect {
    my ($self) = @_;
    my @variables;
    my $statements = $self->_makeDB;
    if (@{$self->{RESULTS}} == 1 && @{$self->{RESULTS}[0]} == 0 && 0) {
	return ([], []);
    }
    for (my $i = 0; $i < @{$self->{RESULTS}}; $i++) {
	my $result = $self->{RESULTS}[$i];
	my @row;
	foreach my $select (@{$self->{SELECTS}}) {
	    if (ref $select && $select->isa('W3C::Rdf::AtomBase')) {
		push (@row, $select);
	    } elsif (!defined $select) {
		push (@row, undef);	# leave column unbound
	    } else {
		push (@row, $result->[$select]);
	    }
	}
	push (@variables, \@row);
	$statements->copyTriples($self->{STATEMENTS}[$i]);
    }
    $self->{SELECTS} = [];
#    return (\@variables, $statements);
    return (\@variables, [@{$self->{STATEMENTS}}]);
}

sub debugCheck {
    my ($self) = @_;
    for (my $i = 0; $i < @{$self->{STATEMENTS}}; $i++) {
	$self->{STATEMENTS}[$i]->debugCheck;
    }
}

sub toString {
    my ($self, %flags) = @_;
    require W3C::Util::TableRenderer;
    my $df = sub {
	eval {
	    if (ref $_) {
		if ($_ == $Value_NULL) {
		    $_ = 'NULL';
		} elsif ($_->can('toString')) {
		    $_ = $_->toString(%flags);
		} elsif ($_->can('toString')) {
		    $_ = $_->toString(%flags);
		}
	    }
	};  if ($@) {if (my $ex = &catch('W3C::Util::Exception')) {
	    $_ = $ex->toString(%flags);
	} else {
	    $_ = new W3C::Util::PerlException()->toString(%flags);
	}}
    };
    my $renderer = new W3C::Util::TableRenderer({%flags, -dataFilter => $df, 
						 -rowDemarks => ['[', '|', ']']});
    $renderer->addHeaders([$self->{INDEX_VAR}]);

    my $results =$self->{RESULTS};
    if (@$results == 0) {
	$renderer->addData([['-no solutions-']]);
    } elsif (@$results == 1 && @{$results->[0]} == 0) {
	$renderer->addData([['-initialized-']]);
    } else {
	for (my $row = 0; $row < @$results; $row++) {
	    $renderer->addData($results->[$row]);
	    if (!$flags{-noProofs}) {
		my $proofStr = join ("\n", map {$_->toString} $self->{STATEMENTS}->[$row]->getTriples());
		$renderer->addRow($proofStr);
	    }
	}
    }
    return $renderer->toString(%flags);
}
# <ITERATOR>
sub elements {
    my ($self) = @_;
    my $ret = new W3C::Rdf::ResultSetIterator($self);
    push (@{$self->{ITERATORS}}, $ret);
    return $ret;
}
# </ITERATOR>

sub markUsed {
    my ($self) = @_;
    $self->{UNUSED} = 0;
}

package W3C::Rdf::ChildResultSet;
@W3C::Rdf::ChildResultSet::ISA = qw(W3C::Rdf::ResultSet);
use W3C::Util::Exception;
use W3C::Rdf::Atoms qw($Value_NULL);
#use W3C::Util::SExpr;

sub new {
    my ($proto, $parent, @parms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new(@parms);
#    $self->{-sources} ||= [];
    $self->{PARENT} = $parent;
    # Overloaded values from parent constructor.

    $self->{VAR_INDEX} = $parent->{VAR_INDEX};
    $self->{INDEX_VAR} = $parent->{INDEX_VAR};
    return $self;
}

package W3C::Rdf::ResultSet;

1;

__END__

=head1 NAME

W3C::Rdf::ResultSet - query state (results and proofs)

=head1 SYNOPSIS

  use W3C::Rdf::Atoms;
  use W3C::Rdf::RdfDB;
  use W3C::Util::NamespaceHandler;
  use W3C::Rdf::Algae2;

  my $atoms = new W3C::Rdf::Atoms();
  my $rs = new W3C::Rdf::ResultSet(-atomDictionary => $atoms);
  $rs->addSelect('firstName');
  for (my $e = $rs->elements; $e->hasMoreElements;) {
      my $row = $e->nextElement;
      my @matches = &getMatches($row);
      foreach my $match (@matches) {
         my $newRow = $row->duplicate();
         $newRow->addProofs([$match]);
         for (my $i = 0; $i < $self->arity; $i++) {
            $newRow->set($resultSet->getIndex('varName'), $match->getPosition($i));
         }
      }
      $row->eliminate;
  }
  my ($nodes, $statements) = $rs->gatherCollect();

=head1 DESCRIPTION

The ResultSet stores the state for a query. It associates variables with field
positions in a set of tuples. For each tuple, there is a corresponding proof,
in the form of an RdfDB containing statements supporting that tuple.

The ResultSet tuples are traversed with an Enumerator which is
returned by the C<elemens()> method.  If the ResultSet is fresh (no
tuple manipulations have occured), there is one tuple with no elements
in it.

Tuple set manipulation occurs through the maniplation of individual tuples
(see TUPLE MANIPULATION).

This module is part of the W3C::Rdf CPAN module.

=head1 CONSTRUCTOR

=head2 new ( ATOM [, OPTS] )

ATOM is an instance of an W3C::Rdf::Atoms.

Returns a new ResultSet object

=head1 METHODS

=head2 elements ()

Get an enumerator for walking the tuples in the result set.

=head2 getIndex ( VAR )

Return the index for an already known variable or assign a new index if the
ResultSet has not yet encountered the variable.

=head2 addSelect ( VAR )

Add a variable to the list to be selected. See L<gatherCollect>

=head2 gatherCollect ()

Copy data from the solution tuples into a report. For each tuple, the fields
corresponding to variables that have been selected by the C<addSelect> method
witll be copied into the report. In addition, RdfDBs containing the proofs for
each tuple are copied into the report. The report is in the form:

  ( array of tuples, array of proofs )

=head1 TUPLE MANIPULATION

Tuple set manipulation occurs through the maniplation of individual tuples.

=head2 duplicate ()

Make the current tuple into two rows. Return the latter of the two.

=head2 addProofs ( STATEMENTS )

Associate more proofs with a given tuple.

=head2 set ( INDEX, VALUE )

Set a field identified by INDEX to VALUE.

=head2 eliminate ()

Remove this row from the solution set.

=head1 AUTHOR

Eric Prud'hommeaux <eric@w3.org>

=head1 SEE ALSO

L<W3C::Rdf::RdfDB>
L<W3C::Rdf::Atoms>

=cut
