-- makeOrderTracking.sql -- Create tables for storing generic RDF statements.
-- $Id: makeGenericTripleStore.sql,v 1.1 2004/06/17 15:55:30 eric Exp $

-- These instructions assume you want a database called rdf. If you
-- want a different database name, adjust accordingly.
-- In MySQL, you can create this with:
--   mysqladmin create rdf

-- MySQL invocation:
--   mysql -u root rdf < makeGenericTripleStore.sql

-- Create generic triplestore tables:

CREATE TABLE __Big__ (id int(10) unsigned NOT NULL auto_increment,
		      hash varchar(32) NOT NULL default '',
		      str text NOT NULL,
		      PRIMARY KEY  (id),
		      UNIQUE KEY u_hash (hash)) TYPE=MyISAM;

CREATE TABLE __Nodes__ (id int(10) unsigned NOT NULL auto_increment,
		        type enum('uri','literal','bnode','table') default NULL,
		        hash varchar(32) NOT NULL default '',
		        str varchar(255) NOT NULL default '',
		        big int(10) unsigned default NULL,
		        attribOrDT int(10) unsigned NOT NULL default '0',
		        PRIMARY KEY  (id),
		        UNIQUE KEY u_hash (type,hash),
		        UNIQUE KEY u_type_str_big_attribOrDT (type,str,big,attribOrDT));

CREATE TABLE __Attributions__ (id int(11) NOT NULL auto_increment,
			       type enum('source','generated','reified') NOT NULL default 'source',
			       doc int(11) unsigned NOT NULL default '0',
			       auth int(11) unsigned default NULL,
			       modified timestamp(14) NOT NULL,
			       created timestamp(14) NOT NULL,
			       PRIMARY KEY  (id),
			       UNIQUE KEY u_type_doc (type,doc));

CREATE TABLE __AttrLists__ (id int(11) NOT NULL auto_increment,
			    listId int(11) NOT NULL default '0',
			    a int(11) NOT NULL default '0',
			    PRIMARY KEY  (id),
			    UNIQUE KEY u_listId_a (listId,a));

CREATE TABLE __Holds__ (id int(10) unsigned NOT NULL auto_increment,
		        p int(10) unsigned NOT NULL default '0',
		        s int(10) unsigned NOT NULL default '0',
		        o int(10) unsigned NOT NULL default '0',
		        r int(10) unsigned NOT NULL default '0',
		        a int(10) unsigned NOT NULL default '0',
		        PRIMARY KEY  (id),
		        UNIQUE KEY u_ (p,s,o,r,a));

-- Handy triple dump query:
-- SELECT s.str,p.str,o.str from __Holds__ AS h, __Nodes__ AS s, __Nodes__ AS p, __Nodes__ AS o WHERE s.id=h.s AND p.id=h.p AND o.id=h.o;

